"""
Python Reddit API Wrapper.

PRAW, an acronym for "Python Reddit API Wrapper", is a python package that
allows for simple access to reddit's API. PRAW aims to be as easy to use as
possible and is designed to follow all of reddit's API rules. You have to give
a useragent, everything else is handled by PRAW so you needn't worry about
violating them.

More information about PRAW can be found at https://github.com/praw-dev/praw
"""

from .const import __version__  # NOQA
from .reddit import Reddit  # NOQA



"""Provide the PRAW models."""
from .auth import Auth  # NOQA
from .front import Front  # NOQA
from .helpers import LiveHelper, MultiredditHelper, SubredditHelper  # NOQA
from .inbox import Inbox  # NOQA
from .list.redditor import RedditorList  # NOQA
from .listing.generator import ListingGenerator  # NOQA
from .listing.listing import Listing  # NOQA
from .reddit.comment import Comment  # NOQA
from .reddit.live import LiveThread  # NOQA
from .reddit.message import Message, SubredditMessage  # NOQA
from .reddit.modaction import ModAction  # NOQA
from .reddit.more import MoreComments  # NOQA
from .reddit.multi import Multireddit  # NOQA
from .reddit.redditor import Redditor  # NOQA
from .reddit.submission import Submission  # NOQA
from .reddit.subreddit import Subreddit  # NOQA
from .reddit.wikipage import WikiPage  # NOQA
from .subreddits import Subreddits  # NOQA
from .user import User  # NOQA

__all__ = ('Auth', 'Comment', 'Front', 'Inbox', 'Listing', 'ListingGenerator',
           'LiveHelper', 'LiveThread', 'Message', 'ModAction', 'MoreComments',
           'Multireddit', 'MultiredditHelper', 'Redditor', 'RedditorList',
           'Submission', 'Subreddit', 'SubredditHelper', 'SubredditMessage',
           'Subreddits', 'User', 'WikiPage')



"""Provide the ReportableMixin class."""
from ....const import API_PATH


class ReportableMixin(object):
    """Interface for RedditBase classes that can be reported."""

    def report(self, reason):
        """Report this object to the moderators of its subreddit.

        :param reason: The reason for reporting.

        """
        self._reddit.post(API_PATH['report'],
                          data={'id': self.fullname, 'reason': reason})



"""Provide CommentForest for Submission comments."""
from heapq import heappop, heappush

from .reddit.more import MoreComments


class CommentForest(object):
    """A forest of comments starts with  multiple top-level comments.

    Each of these comments can be a tree of replies.

    """

    @staticmethod
    def _extract_more_comments(tree):
        """Return a list of MoreComments objects removed from tree."""
        more_comments = []
        queue = [(None, x) for x in tree]
        while len(queue) > 0:
            parent, comment = queue.pop(0)
            if isinstance(comment, MoreComments):
                heappush(more_comments, comment)
                if parent:
                    parent.replies.remove(comment)
                else:
                    tree.remove(comment)
            else:
                for item in comment.replies:
                    queue.append((comment, item))
        return more_comments

    def __getitem__(self, index):
        """Return the comment at position ``index`` in the list."""
        return self._comments[index]

    def __init__(self, submission):
        """Initialize a CommentForest instance.

        :param submission: An instance of :class:`~.Subreddit` that is the
            parent of the comments.
        """
        self._comments = None
        self._comments_by_id = {}
        self._submission = submission

    def __len__(self):
        """Return the number of top-level comments in the forest."""
        return len(self._comments)

    def _insert_comment(self, comment):
        assert comment.name not in self._comments_by_id
        comment.submission = self._submission
        if comment.is_root:
            self._comments.append(comment)
        else:
            assert comment.parent_id in self._comments_by_id
            self._comments_by_id[comment.parent_id].replies.append(comment)

    def _update(self, comments):
        self._comments = comments
        for comment in comments:
            comment.submission = self._submission

    def list(self):
        """Return a flattened list of all Comments.

        This list may contain MoreComments instances if ``replace_more`` was
        not called first.

        """
        comments = []
        queue = list(self)
        while queue:
            comment = queue.pop(0)
            comments.append(comment)
            if not isinstance(comment, MoreComments):
                queue.extend(comment.replies)
        return comments

    def replace_more(self, limit=32, threshold=0):
        """Update the comment forest by resolving instances of MoreComments.

        :param limit: The maximum number of MoreComments instances to
            replace. Each replacement requires 1 API request. Set to ``None``
            to have no limit, or to ``0`` to remove all MoreComments instances
            without additional requests (Default: 32).
        :param threshold: The minimum number of children comments a
            MoreComments instance must have in order to be
            replaced. MoreComments instances that represent "continue this
            thread" links unfortunately appear to have 0 children. (Default:
            0).
        :returns: A list of MoreComments instances that were not replaced.

        """
        remaining = limit
        more_comments = self._extract_more_comments(self._comments)
        skipped = []

        # Fetch largest more_comments until reaching the limit or the threshold
        while (remaining is None or remaining > 0) and more_comments:
            item = heappop(more_comments)
            if item.count < threshold:
                skipped.append(item)
                continue

            new_comments = item.comments(update=False)
            if remaining is not None:
                remaining -= 1

            # Re-add new MoreComment objects to the heap of more_comments
            for more in self._extract_more_comments(new_comments):
                more.submission = self._submission
                heappush(more_comments, more)
            # Insert the new comments into the tree
            for comment in new_comments:
                self._insert_comment(comment)

        return more_comments + skipped



"""Provide the SubmissionListingMixin class."""
from ....const import API_PATH
from ..generator import ListingGenerator
from .base import BaseListingMixin
from .gilded import GildedListingMixin


class SubmissionListingMixin(BaseListingMixin, GildedListingMixin):
    """Adds additional methods pertaining to Submission instances."""

    def duplicates(self, **generator_kwargs):
        """Return a ListingGenerator for the submission's duplicates.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        url = API_PATH['duplicates'].format(submission_id=self.id)
        return ListingGenerator(self._reddit, url, **generator_kwargs)



"""Package providing reddit class mixins."""
from .editable import EditableMixin  # NOQA
from .gildable import GildableMixin  # NOQA
from .inboxable import InboxableMixin  # NOQA
from .messageable import MessageableMixin  # NOQA
from .replyable import ReplyableMixin  # NOQA
from .reportable import ReportableMixin  # NOQA
from .savable import SavableMixin  # NOQA
from .votable import VotableMixin  # NOQA


class UserContentMixin(EditableMixin, GildableMixin, ReplyableMixin,
                       ReportableMixin, SavableMixin, VotableMixin):
    """A convenience mixin that applies to both Comments and Submissions."""



"""Provides the User class."""
from ..const import API_PATH
from .base import PRAWBase
from .listing.generator import ListingGenerator
from .reddit.redditor import Redditor


class User(PRAWBase):
    """The user class provides methods for the currently authenticated user."""

    def blocked(self):
        """Return a RedditorList of blocked Redditors."""
        return self._reddit.get(API_PATH['blocked'])

    def contributor_subreddits(self, **generator_kwargs):
        """Return a ListingGenerator of subreddits user is a contributor of.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, API_PATH['my_contributor'],
                                **generator_kwargs)

    def friends(self):
        """Return a RedditorList of friends."""
        return self._reddit.get(API_PATH['friends'])

    def me(self):  # pylint: disable=invalid-name
        """Return a Redditor instance for the authenticated user."""
        user_data = self._reddit.get(API_PATH['me'])
        return Redditor(self._reddit, _data=user_data)

    def moderator_subreddits(self, **generator_kwargs):
        """Return a ListingGenerator of subreddits the user is a moderator of.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, API_PATH['my_moderator'],
                                **generator_kwargs)

    def multireddits(self):
        """Return a list of multireddits belonging to the user."""
        return self._reddit.get(API_PATH['my_multireddits'])

    def subreddits(self, **generator_kwargs):
        """Return a ListingGenerator of subreddits the user is subscribed to.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, API_PATH['my_subreddits'],
                                **generator_kwargs)



import pickle

import pytest
from praw.models import Subreddit, WikiPage

from ... import UnitTest


class TestSubredit(UnitTest):
    def test_equality(self):
        subreddit1 = Subreddit(self.reddit,
                               _data={'display_name': 'dummy1', 'n': 1})
        subreddit2 = Subreddit(self.reddit,
                               _data={'display_name': 'Dummy1', 'n': 2})
        subreddit3 = Subreddit(self.reddit,
                               _data={'display_name': 'dummy3', 'n': 2})
        assert subreddit1 == subreddit1
        assert subreddit2 == subreddit2
        assert subreddit3 == subreddit3
        assert subreddit1 == subreddit2
        assert subreddit2 != subreddit3
        assert subreddit1 != subreddit3
        assert 'dummy1' == subreddit1
        assert subreddit2 == 'dummy1'

    def test_construct_failure(self):
        message = 'Either `display_name` or `_data` must be provided.'
        with pytest.raises(TypeError) as excinfo:
            Subreddit(self.reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Subreddit(self.reddit, 'dummy', {'id': 'dummy'})
        assert str(excinfo.value) == message

    def test_fullname(self):
        subreddit = Subreddit(self.reddit, _data={'display_name': 'name',
                                                  'id': 'dummy'})
        assert subreddit.fullname == 't5_dummy'

    def test_hash(self):
        subreddit1 = Subreddit(self.reddit,
                               _data={'display_name': 'dummy1', 'n': 1})
        subreddit2 = Subreddit(self.reddit,
                               _data={'display_name': 'Dummy1', 'n': 2})
        subreddit3 = Subreddit(self.reddit,
                               _data={'display_name': 'dummy3', 'n': 2})
        assert hash(subreddit1) == hash(subreddit1)
        assert hash(subreddit2) == hash(subreddit2)
        assert hash(subreddit3) == hash(subreddit3)
        assert hash(subreddit1) == hash(subreddit2)
        assert hash(subreddit2) != hash(subreddit3)
        assert hash(subreddit1) != hash(subreddit3)

    def test_pickle(self):
        subreddit = Subreddit(self.reddit, _data={'display_name': 'name',
                                                  'id': 'dummy'})
        for level in range(pickle.HIGHEST_PROTOCOL + 1):
            other = pickle.loads(pickle.dumps(subreddit, protocol=level))
            assert subreddit == other

    def test_repr(self):
        subreddit = Subreddit(self.reddit, display_name='name')
        assert repr(subreddit) == 'Subreddit(display_name=\'name\')'

    def test_str(self):
        subreddit = Subreddit(self.reddit, _data={'display_name': 'name',
                                                  'id': 'dummy'})
        assert str(subreddit) == 'name'

    def test_submit_failure(self):
        message = 'Either `selftext` or `url` must be provided.'
        subreddit = Subreddit(self.reddit, display_name='name')

        with pytest.raises(TypeError) as excinfo:
            subreddit.submit('Cool title')
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            subreddit.submit('Cool title', selftext='a', url='b')
        assert str(excinfo.value) == message


class TestSubreditWiki(UnitTest):
    def test__getitem(self):
        subreddit = Subreddit(self.reddit, display_name='name')
        wikipage = subreddit.wiki['foo']
        assert isinstance(wikipage, WikiPage)
        assert 'foo' == wikipage.name



"""Provide the Listing class."""
from ..base import PRAWBase


class Listing(PRAWBase):
    """A listing is a collection of RedditBase instances."""

    CHILD_ATTRIBUTE = 'children'

    def __len__(self):
        """Return the number of items in the Listing."""
        return len(getattr(self, self.CHILD_ATTRIBUTE))

    def __getitem__(self, index):
        """Return the item at position index in the list."""
        return getattr(self, self.CHILD_ATTRIBUTE)[index]

    def __setattr__(self, attribute, value):
        """Objectify the CHILD_ATTRIBUTE attribute."""
        if attribute == self.CHILD_ATTRIBUTE:
            value = self._reddit._objector.objectify(value)
        super(Listing, self).__setattr__(attribute, value)


class FlairListing(Listing):
    """Special Listing for handling flair lists."""

    CHILD_ATTRIBUTE = 'users'

    @property
    def after(self):
        """Return the next attribute or None."""
        if hasattr(self, 'next'):
            return self.next  # pylint: disable=no-member
        return None



"""Provide the LiveThread class."""
from ...const import API_PATH
from .base import RedditBase


class LiveThread(RedditBase):
    """An individual LiveThread object."""

    STR_FIELD = 'id'

    def _info_path(self):
        return API_PATH['liveabout'].format(id=self.id)



"""Test praw.models.redditor."""
import mock
import pytest
from prawcore import BadRequest, Forbidden

from ... import IntegrationTest


class TestRedditor(IntegrationTest):
    FRIEND = 'PyAPITestUser3'

    def test_friend(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestRedditor.test_friend'):
            self.reddit.redditor(self.FRIEND.lower()).friend()

    def test_friend__with_note__no_gold(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditor.test_friend__with_note__no_gold'):
            with pytest.raises(BadRequest) as excinfo:
                self.reddit.redditor(self.FRIEND.lower()).friend(note='praw')
            assert 'GOLD_REQUIRED' == excinfo.value.response.json()['reason']

    @mock.patch('time.sleep', return_value=None)
    def test_friend_info(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditor.test_friend_info'):
            redditor = self.reddit.redditor(self.FRIEND).friend_info()
            assert self.FRIEND == redditor
            assert 'date' in redditor.__dict__
            assert 'created_utc' not in redditor.__dict__
            assert hasattr(redditor, 'created_utc')

    def test_gild__no_creddits(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestRedditor.test_gild__no_creddits'):
            with pytest.raises(BadRequest) as excinfo:
                self.reddit.redditor('subreddit_stats').gild()
            reason = excinfo.value.response.json()['reason']
            assert 'INSUFFICIENT_CREDDITS' == reason

    @mock.patch('time.sleep', return_value=None)
    def test_message(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestRedditor.test_message'):
            redditor = self.reddit.redditor('subreddit_stats')
            redditor.message('PRAW test', 'This is a test from PRAW')

    @mock.patch('time.sleep', return_value=None)
    def test_message_from_subreddit(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditor.test_message_from_subreddit'):
            redditor = self.reddit.redditor('subreddit_stats')
            redditor.message('PRAW test', 'This is a test from PRAW',
                             from_subreddit=pytest.placeholders.test_subreddit)

    def test_multireddits(self):
        redditor = self.reddit.redditor('kjoneslol')
        with self.recorder.use_cassette('TestRedditor.test_multireddits'):
            for multireddit in redditor.multireddits():
                if 'sfwpornnetwork' == multireddit.name:
                    break
            else:
                assert False, 'sfwpornnetwork not found in multireddits'

    @mock.patch('time.sleep', return_value=None)
    def test_unfriend(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestRedditor.test_unfriend'):
            redditor = self.reddit.user.friends()[0]
            assert redditor.unfriend() is None

    @mock.patch('time.sleep', return_value=None)
    def test_unblock(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestRedditor.test_unblock'):
            redditor = self.reddit.user.blocked()[0]
            redditor.unblock()


class TestRedditorListings(IntegrationTest):
    def test_comments__controversial(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_comments__controversial'):
            redditor = self.reddit.redditor('spez')
            comments = list(redditor.comments.controversial())
        assert len(comments) == 100

    def test_comments__hot(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_comments__hot'):
            redditor = self.reddit.redditor('spez')
            comments = list(redditor.comments.hot())
        assert len(comments) == 100

    def test_comments__new(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_comments__new'):
            redditor = self.reddit.redditor('spez')
            comments = list(redditor.comments.new())
        assert len(comments) == 100

    def test_comments__top(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_comments__top'):
            redditor = self.reddit.redditor('spez')
            comments = list(redditor.comments.top())
        assert len(comments) == 100

    def test_controversial(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_controversial'):
            redditor = self.reddit.redditor('spez')
            submissions = list(redditor.controversial())
        assert len(submissions) == 100

    def test_downvoted(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_downvoted'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            submissions = list(redditor.downvoted())
        assert len(submissions) > 0

    def test_downvoted__in_read_only_mode(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_downvoted__in_read_only_mode'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            with pytest.raises(Forbidden):
                list(redditor.downvoted())

    def test_downvoted__other_user(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_downvoted__other_user'):
            redditor = self.reddit.redditor('spez')
            with pytest.raises(Forbidden):
                list(redditor.downvoted())

    def test_gilded(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_gilded'):
            redditor = self.reddit.redditor('spez')
            submissions = list(redditor.gilded(limit=50))
        assert len(submissions) == 50

    def test_gildings(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_gildings'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            submissions = list(redditor.gildings())
        assert isinstance(submissions, list)

    def test_gildings__in_read_only_mode(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_gildings__in_read_only_mode'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            with pytest.raises(Forbidden):
                list(redditor.gildings())

    def test_gildings__other_user(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_gildings__other_user'):
            redditor = self.reddit.redditor('spez')
            with pytest.raises(Forbidden):
                list(redditor.gildings())

    def test_hidden(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_hidden'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            submissions = list(redditor.hidden())
        assert len(submissions) > 0

    def test_hidden__in_read_only_mode(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_hidden__in_read_only_mode'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            with pytest.raises(Forbidden):
                list(redditor.hidden())

    def test_hidden__other_user(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_hidden__other_user'):
            redditor = self.reddit.redditor('spez')
            with pytest.raises(Forbidden):
                list(redditor.hidden())

    def test_hot(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_hot'):
            redditor = self.reddit.redditor('spez')
            submissions = list(redditor.hot())
        assert len(submissions) == 100

    def test_new(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_new'):
            redditor = self.reddit.redditor('spez')
            submissions = list(redditor.new())
        assert len(submissions) == 100

    def test_saved(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_saved'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            submissions = list(redditor.saved())
        assert len(submissions) > 0

    def test_saved__in_read_only_mode(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_saved__in_read_only_mode'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            with pytest.raises(Forbidden):
                list(redditor.saved())

    def test_saved__other_user(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_saved__other_user'):
            redditor = self.reddit.redditor('spez')
            with pytest.raises(Forbidden):
                list(redditor.saved())

    def test_submissions__controversial(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_submissions__controversial'):
            redditor = self.reddit.redditor('spladug')
            submissions = list(redditor.submissions.controversial())
        assert len(submissions) == 100

    def test_submissions__hot(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_submissions__hot'):
            redditor = self.reddit.redditor('spez')
            submissions = list(redditor.submissions.hot())
        assert len(submissions) == 100

    def test_submissions__new(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_submissions__new'):
            redditor = self.reddit.redditor('spez')
            submissions = list(redditor.submissions.new())
        assert len(submissions) == 100

    def test_submissions__top(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_submissions__top'):
            redditor = self.reddit.redditor('spladug')
            submissions = list(redditor.submissions.top())
        assert len(submissions) == 100

    def test_top(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_top'):
            redditor = self.reddit.redditor('spez')
            submissions = list(redditor.top())
        assert len(submissions) == 100

    def test_upvoted(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_upvoted'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            submissions = list(redditor.upvoted())
        assert len(submissions) > 0

    def test_upvoted__in_read_only_mode(self):
        with self.recorder.use_cassette(
                'TestRedditorListings.test_upvoted__in_read_only_mode'):
            redditor = self.reddit.redditor(self.reddit.config.username)
            with pytest.raises(Forbidden):
                list(redditor.upvoted())

    def test_upvoted__other_user(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestRedditorListings.test_upvoted__other_user'):
            redditor = self.reddit.redditor('spez')
            with pytest.raises(Forbidden):
                list(redditor.upvoted())



"""Provide the Multireddit class."""
from json import dumps
import re

from ...const import API_PATH
from ..listing.mixins import SubredditListingMixin
from .base import RedditBase
from .redditor import Redditor
from .subreddit import Subreddit


class Multireddit(RedditBase, SubredditListingMixin):
    """A class for users' Multireddits."""

    STR_FIELD = 'path'
    RE_INVALID = re.compile(r'(\s|\W|_)+', re.UNICODE)

    @staticmethod
    def sluggify(title):
        """Return a slug version of the title.

        :param title: The title to make a slug of.

        Adapted from reddit's utils.py.

        """
        title = Multireddit.RE_INVALID.sub('_', title).strip('_').lower()
        if len(title) > 21:  # truncate to nearest word
            title = title[:21]
            last_word = title.rfind('_')
            if last_word > 0:
                title = title[:last_word]
        return title or "_"

    def __init__(self, reddit, _data):
        """Construct an instance of the Multireddit object."""
        self.path = None
        super(Multireddit, self).__init__(reddit, _data)
        self._author = Redditor(reddit, self.path.split('/', 3)[2])
        self._path = API_PATH['multireddit'].format(
            multi=self.name, user=self._author)
        self.path = '/' + self._path[:-1]  # Prevent requests for path
        if 'subreddits' in self.__dict__:
            self.subreddits = [Subreddit(reddit, x['name'])
                               for x in self.subreddits]

    def _info_path(self):
        return API_PATH['multireddit_api'].format(
            multi=self.name, user=self._author)

    def add(self, subreddit):
        """Add a subreddit to this multireddit.

        :param subreddit: The subreddit to add to this multi.

        """
        url = API_PATH['multireddit_update'].format(
            multi=self.name, user=self._author, subreddit=subreddit)
        self._reddit.request(
            'put', url, data={'model': dumps({'name': str(subreddit)})})
        self._reset_attributes('subreddits')

    def copy(self, display_name=None):
        """Copy this multireddit and return the new multireddit.

        :param display_name: (optional) The display name for the copied
            multireddit. Reddit will generate the ``name`` field from this
            display name. When not provided the copy will use the same display
            name and name as this multireddit.

        """
        if display_name:
            name = self.sluggify(display_name)
        else:
            display_name = self.display_name
            name = self.name
        data = {'display_name': display_name, 'from': self.path,
                'to': API_PATH['multireddit'].format(
                    multi=name, user=self._reddit.user.me())}
        return self._reddit.post(API_PATH['multireddit_copy'], data=data)

    def delete(self):
        """Delete this multireddit."""
        self._reddit.request('delete', self._info_path())

    def remove(self, subreddit):
        """Remove a subreddit from this multireddit.

        :param subreddit: The subreddit to remove from this multi.

        """
        url = API_PATH['multireddit_update'].format(
            multi=self.name, user=self._author, subreddit=subreddit)
        self._reddit.request(
            'delete', url, data={'model': dumps({'name': str(subreddit)})})
        self._reset_attributes('subreddits')

    def rename(self, display_name):
        """Rename this multireddit.

        :param display_name: The new display name for this multireddit. Reddit
            will generate the ``name`` field from this display name.

        """
        data = {'from': self.path, 'display_name': display_name}
        updated = self._reddit.post(API_PATH['multireddit_rename'], data=data)
        self.__dict__.update(updated.__dict__)

    def update(self, **updated_settings):
        """Update this multireddit.

        Keyword arguments are passed for settings that should be updated. They
        can any of:

        :param display_name: The display name for this multireddit.
        :param subreddits: Subreddits for this multireddit.
        :param description_md: Description for this multireddit, formatted in
            markdown.
        :param icon_name: Can be one of: ``art and design``, ``ask``,
            ``books``, ``business``, ``cars``, ``comics``, ``cute animals``,
            ``diy``, ``entertainment``, ``food and drink``, ``funny``,
            ``games``, ``grooming``, ``health``, ``life advice``, ``military``,
            ``models pinup``, ``music``, ``news``, ``philosophy``, ``pictures
            and gifs``, ``science``, ``shopping``, ``sports``, ``style``,
            ``tech``, ``travel``, ``unusual stories``, ``video``, or ``None``.
        :param key_color: RGB hex color code of the form `#FFFFFF`.
        :param visibility: Can be one of: ``hidden``, ``private``, ``public``.
        :param weighting_scheme: Can be one of: ``classic``, ``fresh``.

        """
        if 'subreddits' in updated_settings:
            updated_settings['subreddits'] = [
                {'name': str(sub)} for sub in updated_settings['subreddits']]
        response = self._reddit.request(
            'put', self._info_path(), data={'model': dumps(updated_settings)})
        new = Multireddit(self._reddit, response['data'])
        self.__dict__.update(new.__dict__)



import sys
sys.path.insert(0, "..")

from praw import __version__

autoclass_content = 'both'
copyright = '2014, Bryce Boe'
exclude_patterns = ['_build']
extensions = ['sphinx.ext.autodoc']
html_static_path = ['_static']
html_theme = 'default'
html_use_smartypants = True
htmlhelp_basename = 'PRAW'
master_doc = 'index'
nitpicky = True
project = 'PRAW'
pygments_style = 'sphinx'
release = __version__
source_suffix = '.rst'
suppress_warnings = ['image.nonlocal_uri']
version = '.'.join(__version__.split('.', 2)[:2])



"""Test praw.models.front."""
import pytest

from .. import UnitTest


class TestFront(UnitTest):
    def test_controversial_raises_value_error(self):
        with pytest.raises(ValueError):
            self.reddit.front.controversial('second')

    def test_top_raises_value_error(self):
        with pytest.raises(ValueError):
            self.reddit.front.top('second')



from praw.models import Comment
from prawcore import BadRequest
import mock
import pytest

from ... import IntegrationTest


class TestComment(IntegrationTest):
    def test_attributes(self):
        with self.recorder.use_cassette(
                'TestComment.test_attributes'):
            comment = Comment(self.reddit, 'cklhv0f')
            assert comment.author == 'bboe'
            assert comment.body.startswith('Yes it does.')
            assert not comment.is_root
            assert comment.permalink(fast=True) == '/comments/2gmzqe//cklhv0f'
            assert comment.submission == '2gmzqe'

    @mock.patch('time.sleep', return_value=None)
    def test_block(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_block'):
            comment = None
            for item in self.reddit.inbox.submission_replies():
                if item.author and item.author != pytest.placeholders.username:
                    comment = item
                    break
            else:
                assert False, 'no comment found'
            comment.block()

    def test_clear_vote(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_clear_vote'):
            Comment(self.reddit, 'd1680wu').clear_vote()

    @mock.patch('time.sleep', return_value=None)
    def test_delete(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_delete'):
            comment = Comment(self.reddit, 'd1616q2')
            comment.delete()
            assert comment.author is None
            assert comment.body == '[deleted]'

    def test_downvote(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_downvote'):
            Comment(self.reddit, 'd1680wu').downvote()

    @mock.patch('time.sleep', return_value=None)
    def test_edit(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_edit'):
            comment = Comment(self.reddit, 'd1616q2')
            comment.edit('New text')
            assert comment.body == 'New text'

    def test_gild__no_creddits(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_gild__no_creddits'):
            with pytest.raises(BadRequest) as excinfo:
                Comment(self.reddit, 'd1616q2').gild()
            reason = excinfo.value.response.json()['reason']
            assert 'INSUFFICIENT_CREDDITS' == reason

    @mock.patch('time.sleep', return_value=None)
    def test_mark_read(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_mark_read'):
            comment = next(self.reddit.inbox.unread())
            assert isinstance(comment, Comment)
            comment.mark_read()

    @mock.patch('time.sleep', return_value=None)
    def test_mark_unread(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_mark_unread'):
            comment = next(self.reddit.inbox.comment_replies())
            comment.mark_unread()

    def test_permalink(self):
        with self.recorder.use_cassette(
                'TestComment.test_permalink'):
            comment = Comment(self.reddit, 'cklhv0f')
            assert comment.permalink() == ('/r/redditdev/comments/2gmzqe/'
                                           'praw_https_enabled_praw_testing_'
                                           'needed/cklhv0f')

    def test_reply(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_reply'):
            parent_comment = Comment(self.reddit, 'd1616q2')
            comment = parent_comment.reply('Comment reply')
            assert comment.author == self.reddit.config.username
            assert comment.body == 'Comment reply'
            assert not comment.is_root
            assert comment.parent_id == parent_comment.fullname

    def test_report(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_report'):
            Comment(self.reddit, 'd0335z3').report('custom')

    def test_save(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_save'):
            Comment(self.reddit, 'd1680wu').save('foo')

    def test_unsave(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_unsave'):
            Comment(self.reddit, 'd1680wu').unsave()

    def test_upvote(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestComment.test_upvote'):
            Comment(self.reddit, 'd1680wu').upvote()



"""Provide the WikiPage class."""
from ...const import API_PATH
from .base import RedditBase
from .redditor import Redditor


class WikiPage(RedditBase):
    """An individual WikiPage object."""

    @property
    def mod(self):
        """An instance of :class:`.WikiPageModeration`."""
        if self._mod is None:
            self._mod = WikiPageModeration(self)
        return self._mod

    def __eq__(self, other):
        """Return whether the other instance equals the current."""
        return isinstance(other, self.__class__) and \
            str(self).lower() == str(other).lower()

    def __hash__(self):
        """Return the hash of the current instance."""
        return super(WikiPage, self).__hash__()

    def __init__(self, reddit, subreddit, name, _data=None):
        """Construct an instance of the WikiPage object."""
        self.name = name
        self.subreddit = subreddit
        super(WikiPage, self).__init__(reddit, _data)
        self._mod = None

    def __repr__(self):
        """Return an object initialization representation of the instance."""
        return '{}(subreddit={!r}, name={!r})'.format(
            self.__class__.__name__, self.subreddit, self.name)

    def __str__(self):
        """Return a string representation of the instance."""
        return '{}/{}'.format(self.subreddit, self.name)

    def _fetch(self):
        data = self._reddit.get(self._info_path())['data']
        data['revision_by'] = Redditor(self._reddit,
                                       _data=data['revision_by']['data'])
        self.__dict__.update(data)
        self._fetched = True

    def _info_path(self):
        return API_PATH['wiki_page'].format(subreddit=self.subreddit,
                                            page=self.name)

    def edit(self, content, reason=None, **other_settings):
        """Edit this WikiPage's contents.

        :param content: The updated markdown content of the page.
        :param reason: (Optional) The reason for the revision.
        :param other_settings: Additional keyword arguments to pass.

        """
        other_settings.update({'content': content, 'page': self.name,
                               'reason': reason})
        self._reddit.post(API_PATH['wiki_edit'].format(
            subreddit=self.subreddit), data=other_settings)


class WikiPageModeration(object):
    """Provides a set of moderation functions for a WikiPage."""

    def __init__(self, wikipage):
        """Create a WikiPageModeration instance.

        :param wikipage: The wikipage to moderate.

        """
        self.wikipage = wikipage

    def add(self, redditor):
        """Add an editor to this WikiPage.

        :param redditor: A string or :class:`~.Redditor` instance.

        """
        data = {'page': self.wikipage.name, 'username': str(redditor)}
        url = API_PATH['wiki_page_editor'].format(
            subreddit=self.wikipage.subreddit, method='add')
        self.wikipage._reddit.post(url, data=data)

    def remove(self, redditor):
        """Remove an editor from this WikiPage.

        :param redditor: A string or :class:`~.Redditor` instance.

        """
        data = {'page': self.wikipage.name, 'username': str(redditor)}
        url = API_PATH['wiki_page_editor'].format(
            subreddit=self.wikipage.subreddit, method='del')
        self.wikipage._reddit.post(url, data=data)

    def settings(self):
        """Return the settings for this WikiPage."""
        url = API_PATH['wiki_page_settings'].format(
            subreddit=self.wikipage.subreddit, page=self.wikipage.name)
        return self.wikipage._reddit.get(url)['data']

    def update(self, listed, permlevel, **other_settings):
        """Update the settings for this WikiPage.

        :param listed: (boolean) Show this page on page list.
        :param permlevel: (int) Who can edit this page? (0) use subreddit wiki
            permissions, (1) only approved wiki contributors for this page may
            edit (see `add`), (2) only mods may edit and view
        :param other_settings: Additional keyword arguments to pass.
        :returns: The updated WikiPage settings.

        """
        other_settings.update({'listed': listed, 'permlevel': permlevel})
        url = API_PATH['wiki_page_settings'].format(
            subreddit=self.wikipage.subreddit, page=self.wikipage.name)
        return self.wikipage._reddit.post(url, data=other_settings)['data']



import mock
import pytest
from praw import __version__, Reddit
from praw.exceptions import ClientException

from . import UnitTest


class TestReddit(UnitTest):
    REQUIRED_DUMMY_SETTINGS = {x: 'dummy' for x in
                               ['client_id', 'client_secret', 'user_agent']}

    @mock.patch('praw.reddit.update_check')
    def test_check_for_updates(self, mock_update_check):
        Reddit(check_for_updates='1', **self.REQUIRED_DUMMY_SETTINGS)
        assert Reddit.update_checked
        mock_update_check.assert_called_with('praw', __version__)

    def test_comment(self):
        assert self.reddit.comment('cklfmye').id == 'cklfmye'

    def test_context_manager(self):
        with Reddit(**self.REQUIRED_DUMMY_SETTINGS) as reddit:
            assert not reddit.config.check_for_updates

    def test_multireddit(self):
        assert self.reddit.multireddit('bboe', 'aa').path == '/user/bboe/m/aa'

    def test_read_only__with_script_authenticated_core(self):
        with Reddit(password='dummy', username='dummy',
                    **self.REQUIRED_DUMMY_SETTINGS) as reddit:
            assert not reddit.read_only
            reddit.read_only = True
            assert reddit.read_only
            reddit.read_only = False
            assert not reddit.read_only

    def test_read_only__without_trusted_authenticated_core(self):
        with Reddit(password=None, username=None,
                    **self.REQUIRED_DUMMY_SETTINGS) as reddit:
            assert reddit.read_only
            with pytest.raises(ClientException):
                reddit.read_only = False
            assert reddit.read_only
            reddit.read_only = True
            assert reddit.read_only

    def test_read_only__without_untrusted_authenticated_core(self):
        required_settings = self.REQUIRED_DUMMY_SETTINGS.copy()
        required_settings['client_secret'] = None
        with Reddit(password=None, username=None,
                    **required_settings) as reddit:
            assert reddit.read_only
            with pytest.raises(ClientException):
                reddit.read_only = False
            assert reddit.read_only
            reddit.read_only = True
            assert reddit.read_only

    def test_reddit__missing_required_settings(self):
        for setting in self.REQUIRED_DUMMY_SETTINGS:
            with pytest.raises(ClientException) as excinfo:
                settings = self.REQUIRED_DUMMY_SETTINGS.copy()
                del settings[setting]
                Reddit(**settings)
            assert str(excinfo.value).startswith('Required configuration '
                                                 'setting \'{}\' missing.'
                                                 .format(setting))
            if setting == 'client_secret':
                assert 'set to None' in str(excinfo.value)

    def test_reddit__required_settings_set_to_none(self):
        required_settings = self.REQUIRED_DUMMY_SETTINGS.copy()
        del required_settings['client_secret']
        for setting in required_settings:
            with pytest.raises(ClientException) as excinfo:
                settings = self.REQUIRED_DUMMY_SETTINGS.copy()
                settings[setting] = None
                Reddit(**settings)
            assert str(excinfo.value).startswith('Required configuration '
                                                 'setting \'{}\' missing.'
                                                 .format(setting))

    def test_submission(self):
        assert self.reddit.submission('2gmzqe').id == '2gmzqe'

    def test_subreddit(self):
        assert self.reddit.subreddit('redditdev').display_name == 'redditdev'



"""Provides the code to load PRAW's configuration file `praw.ini`."""
import os
import sys

from six.moves import configparser

from .exceptions import ClientException


class Config(object):
    """A class containing the configuration for a reddit site."""

    CONFIG = None
    CONFIG_NOT_SET = object()  # Represents a config value that is not set.

    @classmethod
    def _load_config(cls):
        """Attempt to load settings from various praw.ini files."""
        config = configparser.RawConfigParser()
        module_dir = os.path.dirname(sys.modules[__name__].__file__)
        if 'APPDATA' in os.environ:  # Windows
            os_config_path = os.environ['APPDATA']
        elif 'XDG_CONFIG_HOME' in os.environ:  # Modern Linux
            os_config_path = os.environ['XDG_CONFIG_HOME']
        elif 'HOME' in os.environ:  # Legacy Linux
            os_config_path = os.path.join(os.environ['HOME'], '.config')
        else:
            os_config_path = None
        locations = [os.path.join(module_dir, 'praw.ini'), 'praw.ini']
        if os_config_path is not None:
            locations.insert(1, os.path.join(os_config_path, 'praw.ini'))
        config.read(locations)
        cls.CONFIG = config

    def __init__(self, site_name, **settings):
        """Initialize a Config instance."""
        def config_boolean(item):
            return item.lower() in {'1', 'yes', 'true', 'on'}

        def fetch_or_not_set(key):
            if key in settings:  # Passed in values have the highest priority
                return raw[key]
            return os.getenv('praw_{}'.format(key)) or raw.get(key) \
                or self.CONFIG_NOT_SET

        if Config.CONFIG is None:
            self._load_config()

        raw = dict(Config.CONFIG.items(site_name), **settings)

        self._short_url = raw.get('short_url') or self.CONFIG_NOT_SET
        self.check_for_updates = config_boolean(
            fetch_or_not_set('check_for_updates'))
        self.client_id = fetch_or_not_set('client_id')
        self.client_secret = fetch_or_not_set('client_secret')
        self.http_proxy = fetch_or_not_set('http_proxy')
        self.https_proxy = fetch_or_not_set('https_proxy')
        self.kinds = {x: raw['{}_kind'.format(x)] for x in
                      ['comment', 'message', 'redditor', 'submission',
                       'subreddit']}
        self.oauth_url = raw['oauth_url']
        self.reddit_url = raw['reddit_url']
        self.redirect_uri = fetch_or_not_set('redirect_uri')
        self.refresh_token = fetch_or_not_set('refresh_token')
        self.password = fetch_or_not_set('password')
        self.user_agent = fetch_or_not_set('user_agent')
        self.username = fetch_or_not_set('username')

    @property
    def short_url(self):
        """Return the short url or raise a ClientException when not set."""
        if self._short_url is self.CONFIG_NOT_SET:
            raise ClientException('No short domain specified.')
        return self._short_url



"""Provides the Objector class."""
from .exceptions import APIException


class Objector(object):
    """The objector builds RedditBase objects."""

    def __init__(self, reddit):
        """Initialize an Objector instance.

        :param reddit: An instance of :class:`~.Reddit`.

        """
        self.parsers = {}
        self._reddit = reddit

    def kind(self, instance):
        """Return the kind from the instance class.

        :param instance: An instance of a subclass of RedditBase.

        """
        for key in self.parsers:
            if isinstance(instance, self.parsers[key]):
                return key

    def objectify(self, data):
        """Create RedditBase objects from data.

        :param data: The structured data.

        """
        if isinstance(data, list):
            return [self.objectify(item) for item in data]
        if 'kind' in data and data['kind'] in self.parsers:
            parser = self.parsers[data['kind']]
            return parser.parse(data['data'], self._reddit)
        elif 'json' in data and 'data' in data['json']:
            if 'things' in data['json']['data']:  # Submission.reply
                return self.objectify(data['json']['data']['things'])
            if 'url' in data['json']['data']:  # Subreddit.submit
                # The URL is the URL to the submission, so it's removed.
                del data['json']['data']['url']
                parser = self.parsers[self._reddit.config.kinds['submission']]
            else:
                parser = self.parsers['LiveUpdateEvent']
            return parser.parse(data['json']['data'], self._reddit)
        elif 'json' in data and 'errors' in data['json']:
            errors = data['json']['errors']
            if len(errors) == 1:
                raise APIException(*errors[0])
            assert len(errors) == 0
        elif isinstance(data, dict) and {'date', 'id', 'name'}.issubset(
                set(data.keys())):
            parser = self.parsers[self._reddit.config.kinds['redditor']]
            return parser.parse(data, self._reddit)
        elif isinstance(data, dict) and 'user' in data:
            parser = self.parsers[self._reddit.config.kinds['redditor']]
            data['user'] = parser.parse({'name': data['user']}, self._reddit)
        return data

    def register(self, kind, cls):
        """Register a class for a given kind.

        :param kind: The kind in the parsed data to map to ``cls``.
        :param cls: A RedditBase class.

        """
        self.parsers[kind] = cls



"""Prepare py.test."""
import os
import socket
import time
from base64 import b64encode
from sys import platform

import betamax
from betamax_serializers import pretty_json


# Prevent calls to sleep
def _sleep(*args):
    raise Exception('Call to sleep')
time.sleep = _sleep


def b64_string(input_string):
    """Return a base64 encoded string (not bytes) from input_string."""
    return b64encode(input_string.encode('utf-8')).decode('utf-8')


def env_default(key):
    """Return environment variable or placeholder string."""
    return os.environ.get('prawtest_{}'.format(key),
                          'placeholder_{}'.format(key))


os.environ['praw_check_for_updates'] = 'False'


placeholders = {x: env_default(x) for x in
                ('client_id client_secret password '
                 'test_subreddit user_agent username').split()}
placeholders['basic_auth'] = b64_string(
    '{}:{}'.format(placeholders['client_id'], placeholders['client_secret']))


betamax.Betamax.register_serializer(pretty_json.PrettyJSONSerializer)
with betamax.Betamax.configure() as config:
    config.cassette_library_dir = 'tests/integration/cassettes'
    config.default_cassette_options['serialize_with'] = 'prettyjson'
    for key, value in placeholders.items():
        config.define_cassette_placeholder('<{}>'.format(key.upper()), value)


def pytest_namespace():
    """Add attributes to pytest in all tests."""
    return {'placeholders': placeholders}


if platform == 'darwin':  # Work around issue with betamax on OS X
    socket.gethostbyname = lambda x: '127.0.0.1'



"""Provide the Submission class."""
from six.moves.urllib.parse import (urljoin,  # pylint: disable=import-error
                                    urlparse)

from ...const import API_PATH
from ...exceptions import ClientException
from ..comment_forest import CommentForest
from ..listing.mixins import SubmissionListingMixin
from .base import RedditBase
from .mixins import UserContentMixin
from .redditor import Redditor
from .subreddit import Subreddit


class Submission(RedditBase, SubmissionListingMixin, UserContentMixin):
    """A class for submissions to reddit."""

    STR_FIELD = 'id'

    @staticmethod
    def id_from_url(url):
        """Return the ID contained within a submission URL.

        :param url: A url to a submission in one of the following formats (http
            urls will also work):
            * https://redd.it/2gmzqe
            * https://reddit.com/comments/2gmzqe/
            * https://www.reddit.com/r/redditdev/comments/2gmzqe/praw_https/

        Raise ``ClientException`` if URL is not a valid submission URL.

        """
        parsed = urlparse(url)
        if not parsed.netloc:
            raise ClientException('Invalid URL: {}'.format(url))

        parts = parsed.path.split('/')
        if 'comments' not in parts:
            submission_id = parts[-1]
        else:
            submission_id = parts[parts.index('comments') + 1]

        if not submission_id.isalnum():
            raise ClientException('Invalid URL: {}'.format(url))
        return submission_id

    @property
    def comments(self):
        """An instance of :class:`.CommentForest`."""
        # This assumes _comments is set so that _fetch is called when it's not.
        return self._comments

    @property
    def flair(self):
        """An instance of :class:`.SubmissionFlair`."""
        if self._flair is None:
            self._flair = SubmissionFlair(self)
        return self._flair

    @property
    def mod(self):
        """An instance of :class:`.SubmissionModeration`."""
        if self._mod is None:
            self._mod = SubmissionModeration(self)
        return self._mod

    def __init__(self, reddit, id=None,  # pylint: disable=redefined-builtin
                 url=None, _data=None):
        """Initialize a Submission instance.

        :param reddit: An instance of :class:`~.Reddit`.
        :param id: A reddit base64 submission ID, e.g., ``2gmzqe``.
        :param url: A URL supported by :meth:`~.id_from_url`.

        Either ``id`` or ``url`` can be provided, but not both.

        """
        if [id, url, _data].count(None) != 2:
            raise TypeError('Exactly one of `id`, `url`, or `_data` must be '
                            'provided.')
        super(Submission, self).__init__(reddit, _data)
        self.comment_limit = 2048
        self.comment_sort = 'best'
        if id is not None:
            self.id = id  # pylint: disable=invalid-name
        elif url is not None:
            self.id = self.id_from_url(url)
        self._flair = self._mod = None

    def __setattr__(self, attribute, value):
        """Objectify author, comments, and subreddit attributes."""
        # pylint: disable=redefined-variable-type
        if attribute == 'author':
            value = Redditor.from_data(self._reddit, value)
        elif attribute == 'subreddit':
            value = Subreddit(self._reddit, value)
        super(Submission, self).__setattr__(attribute, value)

    def _fetch(self):
        other, comments = self._reddit.get(self._info_path(),
                                           params={'limit': self.comment_limit,
                                                   'sort': self.comment_sort})
        other = other.children[0]
        other._comments = CommentForest(self)
        self.__dict__.update(other.__dict__)
        self.comments._update(comments.children)
        self._fetched = True

    def _info_path(self):
        return API_PATH['submission'].format(id=self.id)

    def hide(self):
        """Hide Submission."""
        self._reddit.post(API_PATH['hide'], data={'id': self.fullname})

    @property
    def shortlink(self):
        """Return a shortlink to the submission.

        For example http://redd.it/eorhm is a shortlink for
        https://www.reddit.com/r/announcements/comments/eorhm/reddit_30_less_typing/.

        """
        return urljoin(self._reddit.config.short_url, self.id)

    def unhide(self):
        """Unhide Submission."""
        self._reddit.post(API_PATH['unhide'], data={'id': self.fullname})


class SubmissionFlair(object):
    """Provide a set of functions pertaining to Submission flair."""

    def __init__(self, submission):
        """Create a SubmissionFlair instance.

        :param submission: The submission associated with the flair functions.

        """
        self.submission = submission

    def choices(self):
        """Return list of available flair choices."""
        url = API_PATH['flairselector'].format(
            subreddit=self.submission.subreddit)
        return self.submission._reddit.post(url, data={
            'link': self.submission.fullname})['choices']

    def select(self, flair_template_id, text=None):
        """Select flair for submission.

        :param flair_template_id: The flair template to select. The possible
            ``flair_template_id`` values can be discovered through ``choices``.

        :param text: If the template's ``flair_text_editable`` value is True,
            this value will set a custom text (default: None).

        """
        data = {'flair_template_id': flair_template_id,
                'link': self.submission.fullname, 'text': text}
        url = API_PATH['select_flair'].format(
            subreddit=self.submission.subreddit)
        return self.submission._reddit.post(url, data=data)


class SubmissionModeration(object):
    """Provide a set of functions pertaining to Submission moderation."""

    def __init__(self, submission):
        """Create a SubmissionModeration instance.

        :param submission: The submission to moderate.

        """
        self.submission = submission

    def contest_mode(self, state=True):
        """Set contest mode for the comments of this submission.

        :param state: (boolean) True enables contest mode, False, disables.

        Contest mode have the following effects:
          * The comment thread will default to being sorted randomly.
          * Replies to top-level comments will be hidden behind
            "[show replies]" buttons.
          * Scores will be hidden from non-moderators.
          * Scores accessed through the API (mobile apps, bots) will be
            obscured to "1" for non-moderators.

        """
        self.submission._reddit.post(API_PATH['contest_mode'], data={
            'id': self.submission.fullname, 'state': state})

    def nsfw(self):
        """Mark as not safe for work."""
        self.submission._reddit.post(API_PATH['marknsfw'],
                                     data={'id': self.submission.fullname})

    def sfw(self):
        """Mark as safe for work."""
        self.submission._reddit.post(API_PATH['unmarknsfw'],
                                     data={'id': self.submission.fullname})

    def sticky(self, state=True, bottom=True):
        """Set the submission's sticky state in its subreddit.

        :param state: (boolean) True sets the sticky for the submission, false
            unsets (default: True).
        :param bottom: (boolean) When true, set the submission as the bottom
            sticky. If no top sticky exists, this submission will become the
            top sticky regardless (default: True).

        This submission will replace an existing stickied submission if one
        exists.

        """
        data = {'id': self.submission.fullname, 'state': state}
        if not bottom:
            data['num'] = 1
        return self.submission._reddit.post(API_PATH['sticky_submission'],
                                            data=data)

    def suggested_sort(self, sort='blank'):
        """Set the suggested sort for the comments of the submission.

        :param sort: Can be one of: confidence, top, new, controversial, old,
            random, qa, blank (default: blank).

        """
        self.submission._reddit.post(API_PATH['suggested_sort'], data={
            'id': self.submission.fullname, 'sort': sort})

Subreddit._submission_class = Submission



"""Test praw.models.user."""
from praw.models import Multireddit, Redditor, Subreddit

from .. import IntegrationTest


class TestUser(IntegrationTest):
    def test_blocked(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestUser.test_blocked'):
            blocked = self.reddit.user.blocked()
        assert len(blocked) > 0
        assert all(isinstance(user, Redditor) for user in blocked)

    def test_contributor_subreddits(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestUser.test_contributor_subreddits'):
            count = 0
            for subreddit in self.reddit.user.contributor_subreddits():
                assert isinstance(subreddit, Subreddit)
                count += 1
            assert count > 0

    def test_friends(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestUser.test_friends'):
            friends = self.reddit.user.friends()
        assert len(friends) > 0
        assert all(isinstance(friend, Redditor) for friend in friends)

    def test_me(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestUser.test_me'):
            me = self.reddit.user.me()
        assert isinstance(me, Redditor)

    def test_moderator_subreddits(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestUser.test_moderator_subreddits'):
            count = 0
            for subreddit in self.reddit.user.moderator_subreddits():
                assert isinstance(subreddit, Subreddit)
                count += 1
            assert count > 0

    def test_multireddits(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestUser.test_multireddits'):
            multireddits = self.reddit.user.multireddits()
            assert isinstance(multireddits, list)
            assert multireddits
            assert all(isinstance(x, Multireddit) for x in multireddits)

    def test_subreddits(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestUser.test_subreddits'):
            count = 0
            for subreddit in self.reddit.user.subreddits():
                assert isinstance(subreddit, Subreddit)
                count += 1
            assert count > 0



import pickle

import pytest
from praw.exceptions import ClientException
from praw.models import Submission

from ... import UnitTest


class TestSubmission(UnitTest):
    def test_equality(self):
        submission1 = Submission(self.reddit, _data={'id': 'dummy1', 'n': 1})
        submission2 = Submission(self.reddit, _data={'id': 'Dummy1', 'n': 2})
        submission3 = Submission(self.reddit, _data={'id': 'dummy3', 'n': 2})
        assert submission1 == submission1
        assert submission2 == submission2
        assert submission3 == submission3
        assert submission1 == submission2
        assert submission2 != submission3
        assert submission1 != submission3
        assert 'dummy1' == submission1
        assert submission2 == 'dummy1'

    def test_construct_failure(self):
        message = 'Exactly one of `id`, `url`, or `_data` must be provided.'
        with pytest.raises(TypeError) as excinfo:
            Submission(self.reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Submission(self.reddit, id='dummy', url='dummy')
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Submission(self.reddit, 'dummy', _data={'id': 'dummy'})
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Submission(self.reddit, url='dummy', _data={'id': 'dummy'})
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Submission(self.reddit, 'dummy', 'dummy', {'id': 'dummy'})
        assert str(excinfo.value) == message

    def test_construct_from_url(self):
        assert Submission(self.reddit, url='http://my.it/2gmzqe') == '2gmzqe'

    def test_fullname(self):
        submission = Submission(self.reddit, _data={'id': 'dummy'})
        assert submission.fullname == 't3_dummy'

    def test_hash(self):
        submission1 = Submission(self.reddit, _data={'id': 'dummy1', 'n': 1})
        submission2 = Submission(self.reddit, _data={'id': 'Dummy1', 'n': 2})
        submission3 = Submission(self.reddit, _data={'id': 'dummy3', 'n': 2})
        assert hash(submission1) == hash(submission1)
        assert hash(submission2) == hash(submission2)
        assert hash(submission3) == hash(submission3)
        assert hash(submission1) == hash(submission2)
        assert hash(submission2) != hash(submission3)
        assert hash(submission1) != hash(submission3)

    def test_id_from_url(self):
        urls = ['http://my.it/2gmzqe',
                'https://redd.it/2gmzqe',
                'http://reddit.com/comments/2gmzqe',
                'https://www.reddit.com/r/redditdev/comments/2gmzqe/'
                'praw_https_enabled_praw_testing_needed/']
        for url in urls:
            assert Submission.id_from_url(url) == '2gmzqe', url

    def test_id_from_url__invalid_urls(self):
        urls = ['', '1', '/', 'my.it/2gmzqe',
                'http://my.it/_',
                'https://redd.it/_/',
                'http://reddit.com/comments/_/2gmzqe']
        for url in urls:
            with pytest.raises(ClientException):
                Submission.id_from_url(url)

    def test_pickle(self):
        submission = Submission(self.reddit, _data={'id': 'dummy'})
        for level in range(pickle.HIGHEST_PROTOCOL + 1):
            other = pickle.loads(pickle.dumps(submission, protocol=level))
            assert submission == other

    def test_repr(self):
        submission = Submission(self.reddit, id='2gmzqe')
        assert repr(submission) == 'Submission(id=\'2gmzqe\')'

    def test_str(self):
        submission = Submission(self.reddit, _data={'id': 'dummy'})
        assert str(submission) == 'dummy'

    def test_shortlink(self):
        submission = Submission(self.reddit, _data={'id': 'dummy'})
        assert submission.shortlink == 'https://redd.it/dummy'



"""Provide the SavableMixin class."""
from ....const import API_PATH


class SavableMixin(object):
    """Interface for RedditBase classes that can be saved."""

    def save(self, category=None):
        """Save the object.

        :param category: The category to save to (Default: None).

        """
        self._reddit.post(API_PATH['save'], data={'category': category,
                                                  'id': self.fullname})

    def unsave(self):
        """Unsave the object."""
        self._reddit.post(API_PATH['unsave'], data={'id': self.fullname})



"""PRAW exception classes.

Includes two main exceptions: ``APIExeception`` for when something goes wrong
on the server side, and ``ClientException`` when something goes wrong on the
client side. Both of these classes extend ``PRAWException``.

"""


class PRAWException(Exception):
    """The base PRAW Exception that all other exception classes extend."""


class APIException(PRAWException):
    """Indicate exception that involve responses from reddit's API."""

    def __init__(self, error_type, message, field):
        """Construct an APIException.

        :param error_type: The error type set on reddit's end.
        :param message: The associated message for the error.
        :param field: The input field associated with the error if available.

        """
        error_str = '{}: \'{}\''.format(error_type, message)
        if field:
            error_str += ' on field \'{}\''.format(field)
        super(APIException, self).__init__(error_str)
        self.error_type = error_type
        self.message = message
        self.field = field


class ClientException(PRAWException):
    """Indicate exceptions that don't involve interaction with reddit's API."""



"""Provide the Front class."""
from .listing.mixins import SubredditListingMixin


class Front(SubredditListingMixin):
    """Front is a Listing class that represents the front page."""

    def __init__(self, reddit):
        """Initialize a Front instance."""
        super(Front, self).__init__(reddit, None)
        self._path = '/'



"""Provide the ReplyableMixin class."""
from ....const import API_PATH


class ReplyableMixin(object):
    """Interface for RedditBase classes that can be replied to."""

    def reply(self, body):
        """Reply to the object..

        :param body: The markdown formatted content for a comment.
        :returns: A :class:`~.Comment` object for the newly created comment.

        """
        data = {'text': body, 'thing_id': self.fullname}
        return self._reddit.post(API_PATH['comment'], data=data)[0]



"""Test praw.objector."""
from praw.exceptions import APIException
import pytest

from . import IntegrationTest


class TestObjector(IntegrationTest):
    def test_raise_api_exception(self):
        message = 'USER_REQUIRED: \'Please log in to do that.\''
        with self.recorder.use_cassette(
                'TestObjector.test_raise_api_exception'):
            subreddit = self.reddit.subreddit(
                pytest.placeholders.test_subreddit)
            with pytest.raises(APIException) as excinfo:
                subreddit.mod.approve(self.reddit.submission('4b536h'))
            assert excinfo.value.error_type == 'USER_REQUIRED'
            assert str(excinfo.value) == message



"""Provide the PRAWBase superclass."""


class PRAWBase(object):
    """Superclass for all models in PRAW."""

    @classmethod
    def parse(cls, data, reddit):
        """Return an instance of ``cls`` from ``data``.

        :param data: The structured data.
        :param reddit: An instance of :class:`.Reddit`.

        """
        return cls(reddit, _data=data)

    def __init__(self, reddit, _data):
        """Initialize a PRAWModel instance.

        :param reddit: An instance of :class:`.Reddit`.

        """
        self._reddit = reddit
        if _data:
            for attribute, value in _data.items():
                setattr(self, attribute, value)



"""Provide the Reddit class."""
import os

from six import iteritems
from update_checker import update_check
from prawcore import (DeviceIDAuthorizer, ReadOnlyAuthorizer, Redirect,
                      Requestor, ScriptAuthorizer, TrustedAuthenticator,
                      UntrustedAuthenticator, session)

from .exceptions import ClientException
from .config import Config
from .const import __version__, API_PATH, USER_AGENT_FORMAT
from .objector import Objector
from . import models


class Reddit(object):
    """Provide convenient access to reddit's API."""

    update_checked = False

    @property
    def _next_unique(self):
        value = self._unique_counter
        self._unique_counter += 1
        return value

    @property
    def read_only(self):
        """Return True when using the ReadOnlyAuthorizer."""
        return self._core == self._read_only_core

    @read_only.setter
    def read_only(self, value):
        """Set or unset the use of the ReadOnlyAuthorizer.

        Raise ``ClientException`` when attempting to unset ``read_only`` and
        only the ReadOnlyAuthorizer is available.

        """
        if value:
            self._core = self._read_only_core
        elif self._authorized_core is None:
            raise ClientException('read_only cannot be set as only the '
                                  'ReadOnlyAuthorizer is available.')
        else:
            self._core = self._authorized_core

    def __enter__(self):
        """Handle the context manager open."""
        return self

    def __exit__(self, *_args):
        """Handle the context manager close."""
        pass

    def __init__(self, site_name=None, **config_settings):
        """Initialize a Reddit instance.

        :param site_name: The name of a section in your ``praw.ini`` file from
            which to load settings from. This parameter, in tandem with an
            appropriately configured ``praw.ini``, file is useful if you wish
            to easily save credentials for different applications, or
            communicate with other servers running reddit. If ``site_name`` is
            ``None``, then the site name will be looked for in the environment
            variable praw_site. If it is not found there, the DEFAULT site will
            be used.

        Additional keyword arguments will be used to initialize the ``Config``
        object. This can be used to specify configuration settings during
        instantiation of the ``Reddit`` instance. For more details please see:
        https://praw.readthedocs.org/en/stable/pages/configuration_files.html

        Required settings are:

        * client_id
        * client_secret (for installed applications set this value to ``None``)
        * user_agent

        """
        self._core = self._authorized_core = self._read_only_core = None
        self._objector = None
        self._unique_counter = 0
        self.config = Config(site_name or os.getenv('praw_site') or 'DEFAULT',
                             **config_settings)

        required_message = ('Required configuration setting {!r} missing. \n'
                            'This setting can be provided in a praw.ini file, '
                            'as a keyword argument to the `Reddit` class '
                            'constructor, or as an environment variable.')
        for attribute in ('client_id', 'user_agent'):
            if getattr(self.config, attribute) in (self.config.CONFIG_NOT_SET,
                                                   None):
                raise ClientException(required_message.format(attribute))
        if self.config.client_secret is self.config.CONFIG_NOT_SET:
            raise ClientException(required_message.format('client_secret') +
                                  '\nFor installed applications this value '
                                  'must be set to None via a keyword arugment '
                                  'to the `Reddit` class constructor.')

        self._check_for_update()
        self._prepare_objector()
        self._prepare_prawcore()

        #: An instance of :class:`.Auth`.
        self.auth = models.Auth(self, None)

        #: An instance of :class:`.Front`.
        self.front = models.Front(self)

        #: An instance of :class:`.Inbox`.
        self.inbox = models.Inbox(self, None)

        #: An instance of :class:`.LiveHelper`.
        self.live = models.LiveHelper(self, None)

        #: An instance of :class:`.MultiredditHelper`.
        self.multireddit = models.MultiredditHelper(self, None)

        #: An instance of :class:`.SubredditHelper`.
        self.subreddit = models.SubredditHelper(self, None)

        #: An instance of :class:`.Subreddits`.
        self.subreddits = models.Subreddits(self, None)

        #: An instance of :class:`.User`.
        self.user = models.User(self, None)

    def _check_for_update(self):
        if not Reddit.update_checked and self.config.check_for_updates:
            update_check(__package__, __version__)
            Reddit.update_checked = True

    def _prepare_objector(self):
        self._objector = Objector(self)
        mappings = {self.config.kinds['comment']: models.Comment,
                    self.config.kinds['message']: models.Message,
                    self.config.kinds['redditor']: models.Redditor,
                    self.config.kinds['submission']: models.Submission,
                    self.config.kinds['subreddit']: models.Subreddit,
                    'LabeledMulti': models.Multireddit,
                    'Listing': models.Listing,
                    'LiveUpdateEvent': models.LiveThread,
                    'UserList': models.RedditorList,
                    'modaction': models.ModAction,
                    'more': models.MoreComments}
        for kind, klass in iteritems(mappings):
            self._objector.register(kind, klass)

    def _prepare_prawcore(self):
        requestor = Requestor(USER_AGENT_FORMAT.format(self.config.user_agent),
                              self.config.oauth_url, self.config.reddit_url)
        if self.config.client_secret:
            self._prepare_trusted_prawcore(requestor)
        else:
            self._prepare_untrusted_prawcore(requestor)

    def _prepare_trusted_prawcore(self, requestor):
        authenticator = TrustedAuthenticator(requestor, self.config.client_id,
                                             self.config.client_secret,
                                             self.config.redirect_uri)
        read_only_authorizer = ReadOnlyAuthorizer(authenticator)
        self._read_only_core = session(read_only_authorizer)

        if self.config.username and self.config.password:
            script_authorizer = ScriptAuthorizer(
                authenticator, self.config.username, self.config.password)
            self._core = self._authorized_core = session(script_authorizer)
        else:
            self._core = self._read_only_core

    def _prepare_untrusted_prawcore(self, requestor):
        authenticator = UntrustedAuthenticator(requestor,
                                               self.config.client_id,
                                               self.config.redirect_uri)
        read_only_authorizer = DeviceIDAuthorizer(authenticator)
        self._core = self._read_only_core = session(read_only_authorizer)

    def comment(self, id):  # pylint: disable=invalid-name,redefined-builtin
        """Return a lazy instance of :class:`~.Comment` for ``id``.

        :param id: The ID of the comment.

        """
        return models.Comment(self, id=id)

    def get(self, path, params=None):
        """Return parsed objects returned from a GET request to ``path``.

        :param path: The path to fetch.
        :param params: The query parameters to add to the request (Default:
            None).

        """
        data = self.request('GET', path, params=params)
        return self._objector.objectify(data)

    def post(self, path, data=None, params=None):
        """Return parsed objects returned from a POST request to ``path``.

        :param path: The path to fetch.
        :param data: Dictionary, bytes, or file-like object to send in the body
            of the request.
        :param params: The query parameters to add to the request (Default:
            None).

        """
        data = self.request('POST', path, data=data, params=params)
        return self._objector.objectify(data)

    def random_subreddit(self, nsfw=False):
        """Return a random lazy instance of :class:`~.Subreddit`.

        :param nsfw: Return a random NSFW (not safe for work) subreddit
            (Default: False).

        """
        url = API_PATH['subreddit'].format(subreddit='randnsfw' if nsfw
                                           else 'random')
        path = None
        try:
            self.get(url, params={'unique': self._next_unique})
        except Redirect as redirect:
            path = redirect.path
        return models.Subreddit(self, path.split('/')[2])

    def redditor(self, name):
        """Return a lazy instance of :class:`~.Redditor` for ``name``.

        :param name: The name of the redditor.

        """
        return models.Redditor(self, name)

    def request(self, method, path, params=None, data=None):
        """Return the parsed JSON data returned from a request to URL.

        :param method: The HTTP method (e.g., GET, POST, PUT, DELETE).
        :param path: The path to fetch.
        :param params: The query parameters to add to the request (Default:
            None).
        :param data: Dictionary, bytes, or file-like object to send in the body
            of the request.

        """
        return self._core.request(method, path, params=params, data=data)

    def submission(  # pylint: disable=invalid-name,redefined-builtin
            self, id=None, url=None):
        """Return a lazy instance of :class:`~.Submission`.

        :param id: A reddit base64 submission ID, e.g., ``2gmzqe``.
        :param url: A URL supported by :meth:`.id_from_url`.

        Either ``id`` or ``url`` can be provided, but not both.

        """
        return models.Submission(self, id=id, url=url)



"""Provide the Auth class."""
from prawcore import ImplicitAuthorizer, UntrustedAuthenticator, session

from .base import PRAWBase
from ..exceptions import ClientException


class Auth(PRAWBase):
    """Auth provides an interface to Reddit's authorization."""

    def implicit(self, access_token, expires_in, scope):
        """Set the active authorization to be an implicit authorization.

        :param access_token: The access_token obtained from Reddit's callback.
        :param expires_in: The number of seconds the ``access_token`` is valid
            for. The origin of this value was returned from Reddit's callback.
            You may need to subtract an offset before passing in this number to
            account for a delay between when Reddit prepared the response, and
            when you make this function call.
        :param scope: A space-delimited string of Reddit OAuth2 scope names as
            returned from Reddit's callback.

        Raise ``ClientException`` if ``Reddit`` was initialized for a
        non-installed application type.

        """
        authenticator = self._reddit._read_only_core._authorizer._authenticator
        if not isinstance(authenticator, UntrustedAuthenticator):
            raise ClientException('implicit can only be used with installed '
                                  'apps.')
        implicit_session = session(ImplicitAuthorizer(
            authenticator, access_token, expires_in, scope))
        self._reddit._core = self._reddit._authorized_core = implicit_session

    def url(self, scopes, state, duration='permanent'):
        """Return the URL used out-of-band to grant access to your application.

        :param scopes: A list of OAuth scopes to request authorization for.
        :param state: A string that will be reflected in the callback to
            ``redirect_uri``. This value should be temporarily unique to the
            client for whom the URL was generated for.
        :param duration: (web app only) Either ``permanent`` or ``temporary``
            (default: permanent). ``temporary`` authorizations generate access
            tokens that last only 1 hour. ``permanent`` authorizations
            additionally generate a refresh token that can be indefinitely used
            to generate new hour-long access tokens. This value is ignored for
            installed apps as only temporary tokens can be generated for them.

        """
        authenticator = self._reddit._read_only_core._authorizer._authenticator
        if isinstance(authenticator, UntrustedAuthenticator):
            return authenticator.authorize_url(scopes, state)
        raise ClientException('url is not yet supported for web app')



"""Provide the GildableMixin class."""
from ....const import API_PATH


class GildableMixin(object):
    """Interface for classes that can be gilded."""

    def gild(self):
        """Gild the author of the item."""
        self._reddit.post(API_PATH['gild_thing'].format(
            fullname=self.fullname))



from praw.exceptions import APIException, ClientException, PRAWException


class TestPRAWException(object):
    def test_inheritance(self):
        assert isinstance(PRAWException(), Exception)

    def test_str(self):
        assert str(PRAWException()) == ''
        assert str(PRAWException('foo')) == 'foo'


class TestAPIException(object):
    def test_inheritance(self):
        assert isinstance(APIException(None, None, None), PRAWException)

    def test_str(self):
        exception = APIException('BAD_SOMETHING', 'invalid something',
                                 'some_field')
        assert str(exception) == ('BAD_SOMETHING: \'invalid something\' on '
                                  'field \'some_field\'')


class TestClientException(object):
    def test_inheritance(self):
        assert isinstance(ClientException(), PRAWException)

    def test_str(self):
        assert str(ClientException()) == ''
        assert str(ClientException('error message')) == 'error message'



import pickle

import pytest
from praw.models import Comment

from ... import UnitTest


class TestComment(UnitTest):
    def test_equality(self):
        comment1 = Comment(self.reddit, _data={'id': 'dummy1', 'n': 1})
        comment2 = Comment(self.reddit, _data={'id': 'Dummy1', 'n': 2})
        comment3 = Comment(self.reddit, _data={'id': 'dummy3', 'n': 2})
        assert comment1 == comment1
        assert comment2 == comment2
        assert comment3 == comment3
        assert comment1 == comment2
        assert comment2 != comment3
        assert comment1 != comment3
        assert 'dummy1' == comment1
        assert comment2 == 'dummy1'

    def test_construct_failure(self):
        message = 'Either `id` or `_data` must be provided.'
        with pytest.raises(TypeError) as excinfo:
            Comment(self.reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Comment(self.reddit, 'dummy', _data={'id': 'dummy'})
        assert str(excinfo.value) == message

    def test_hash(self):
        comment1 = Comment(self.reddit, _data={'id': 'dummy1', 'n': 1})
        comment2 = Comment(self.reddit, _data={'id': 'Dummy1', 'n': 2})
        comment3 = Comment(self.reddit, _data={'id': 'dummy3', 'n': 2})
        assert hash(comment1) == hash(comment1)
        assert hash(comment2) == hash(comment2)
        assert hash(comment3) == hash(comment3)
        assert hash(comment1) == hash(comment2)
        assert hash(comment2) != hash(comment3)
        assert hash(comment1) != hash(comment3)

    def test_pickle(self):
        comment = Comment(self.reddit, _data={'id': 'dummy'})
        for level in range(pickle.HIGHEST_PROTOCOL + 1):
            other = pickle.loads(pickle.dumps(comment, protocol=level))
            assert comment == other

    def test_repr(self):
        comment = Comment(self.reddit, id='dummy')
        assert repr(comment) == 'Comment(id=\'dummy\')'

    def test_str(self):
        comment = Comment(self.reddit, _data={'id': 'dummy'})
        assert str(comment) == 'dummy'



"""Provide helper classes used by other models."""


class BoundedSet(object):
    """A set with a maximum size that evicts the oldest items when necessary.

    This class does not implement the complete set interface.
    """

    def __init__(self, max_items):
        """Construct an instance of the BoundedSet."""
        self.max_items = max_items
        self._fifo = []
        self._set = set()

    def __contains__(self, item):
        """Test if the BoundedSet contains item."""
        return item in self._set

    def add(self, item):
        """Add an item to the set discarding the oldest item if necessary."""
        if len(self._set) == self.max_items:
            self._set.remove(self._fifo.pop(0))
        self._fifo.append(item)
        self._set.add(item)


def stream_generator(function):
    """Forever yield new items from ListingGenerators."""
    before_fullname = None
    seen_fullnames = BoundedSet(100)
    without_before_counter = 0
    while True:
        newest_fullname = None
        limit = 100
        if before_fullname is None:
            limit -= without_before_counter
            without_before_counter = (without_before_counter + 1) % 30
        for item in reversed(list(function(
                limit=limit, params={'before': before_fullname}))):
            if item.fullname in seen_fullnames:
                continue
            seen_fullnames.add(item.fullname)
            newest_fullname = item.fullname
            yield item
        before_fullname = newest_fullname



from praw.models import Comment, Submission
from prawcore import BadRequest
import mock
import pytest

from ... import IntegrationTest


class TestSubmission(IntegrationTest):
    def test_comments(self):
        with self.recorder.use_cassette(
                'TestSubmission.test_comments'):
            submission = Submission(self.reddit, '2gmzqe')
            assert len(submission.comments) == 1
            assert isinstance(submission.comments[0], Comment)
            assert isinstance(submission.comments[0].replies[0], Comment)

    def test_clear_vote(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_clear_vote'):
            Submission(self.reddit, '4b536p').clear_vote()

    @mock.patch('time.sleep', return_value=None)
    def test_delete(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_delete'):
            submission = Submission(self.reddit, '4b1tfm')
            submission.delete()
            assert submission.author is None
            assert submission.selftext == '[deleted]'

    def test_downvote(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_downvote'):
            Submission(self.reddit, '4b536p').downvote()

    def test_duplicates(self):
        with self.recorder.use_cassette(
                'TestSubmission.test_duplicates'):
            submission = Submission(self.reddit, 'avj2v')
            assert len(list(submission.duplicates())) > 0

    @mock.patch('time.sleep', return_value=None)
    def test_edit(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_edit'):
            submission = Submission(self.reddit, '4b1tfm')
            submission.edit('New text')
            assert submission.selftext == 'New text'

    def test_gild__no_creddits(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_gild__no_creddits'):
            with pytest.raises(BadRequest) as excinfo:
                Submission(self.reddit, '4b1tfm').gild()
            reason = excinfo.value.response.json()['reason']
            assert 'INSUFFICIENT_CREDDITS' == reason

    def test_hide(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_hide'):
            Submission(self.reddit, '4b1tfm').hide()

    def test_invalid_attribute(self):
        with self.recorder.use_cassette(
                'TestSubmission.test_invalid_attribute'):
            submission = Submission(self.reddit, '2gmzqe')
            with pytest.raises(AttributeError) as excinfo:
                submission.invalid_attribute
        assert excinfo.value.args[0] == ("'Submission' object has no attribute"
                                         " 'invalid_attribute'")

    def test_reply(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_reply'):
            submission = Submission(self.reddit, '4b1tfm')
            comment = submission.reply('Test reply')
            assert comment.author == self.reddit.config.username
            assert comment.body == 'Test reply'
            assert comment.parent_id == submission.fullname

    def test_report(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_report'):
            Submission(self.reddit, '4b536h').report('praw')

    def test_save(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_save'):
            Submission(self.reddit, '4b536p').save()

    def test_unhide(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_unhide'):
            Submission(self.reddit, '4b1tfm').unhide()

    def test_unsave(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_unsave'):
            Submission(self.reddit, '4b536p').unsave()

    def test_upvote(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmission.test_upvote'):
            Submission(self.reddit, '4b536p').upvote()


class TestSubmissionFlair(IntegrationTest):
    @mock.patch('time.sleep', return_value=None)
    def test_choices(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubmissionFlair.test_choices'):
            submission = Submission(self.reddit, '4t4t2h')
            expected = [
                {'flair_text': 'SATISFIED',
                 'flair_template_id': '680f43b8-1fec-11e3-80d1-12313b0b80bc',
                 'flair_text_editable': False, 'flair_position': 'left',
                 'flair_css_class': ''},
                {'flair_text': 'STATS',
                 'flair_template_id': '16cabd0a-a68d-11e5-8349-0e8ff96e6679',
                 'flair_text_editable': False, 'flair_position': 'left',
                 'flair_css_class': ''}]
            assert expected == submission.flair.choices()

    @mock.patch('time.sleep', return_value=None)
    def test_select(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubmissionFlair.test_select'):
            submission = Submission(self.reddit, '4t4t2h')
            submission.flair.select('16cabd0a-a68d-11e5-8349-0e8ff96e6679')


class TestSubmissionModeration(IntegrationTest):
    def test_contest_mode(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_contest_mode'):
            submission = Submission(self.reddit, '4s2idz')
            submission.mod.contest_mode()

    def test_contest_mode__disable(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_contest_mode__disable'):
            submission = Submission(self.reddit, '4s2idz')
            submission.mod.contest_mode(state=False)

    def test_nsfw(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_nsfw'):
            Submission(self.reddit, '4s2idz').mod.nsfw()

    def test_sfw(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_sfw'):
            Submission(self.reddit, '4s2idz').mod.sfw()

    def test_sticky(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_sticky'):
            Submission(self.reddit, '4s2idz').mod.sticky()

    def test_sticky__remove(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_sticky__remove'):
            Submission(self.reddit, '4s2idz').mod.sticky(state=False)

    def test_sticky__top(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_sticky__top'):
            Submission(self.reddit, '4s2idz').mod.sticky(bottom=False)

    def test_suggested_sort(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_suggested_sort'):
            Submission(self.reddit, '4s2idz').mod.suggested_sort(sort='random')

    def test_suggested_sort__clear(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubmissionModeration.test_suggested_sort__clear'):
            Submission(self.reddit, '4s2idz').mod.suggested_sort(sort='blank')



from praw.models import Redditor, WikiPage
from prawcore import NotFound
import pytest

from ... import IntegrationTest


class TestWikiPage(IntegrationTest):
    def test_content_md(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'test')

        with self.recorder.use_cassette('TestWikiPage.test_content_md'):
            assert page.content_md

    def test_edit(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'test')

        self.reddit.read_only = False
        with self.recorder.use_cassette('TestWikiPage.test_edit'):
            page.edit('PRAW updated')

    def test_edit__with_reason(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'test')

        self.reddit.read_only = False
        with self.recorder.use_cassette('TestWikiPage.test_edit__with_reason'):
            page.edit('PRAW updated with reason', reason='PRAW testing')

    def test_invalid_page(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'invalid')

        with self.recorder.use_cassette('TestWikiPage.test_invalid_page'):
            with pytest.raises(NotFound):
                page.content_md

    def test_revision_by(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'test')

        with self.recorder.use_cassette('TestWikiPage.test_revision_by'):
            assert isinstance(page.revision_by, Redditor)


class TestWikiPageModeration(IntegrationTest):
    def test_add(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'test')

        self.reddit.read_only = False
        with self.recorder.use_cassette('TestWikiPageModeration.test_add'):
            page.mod.add('bboe')

    def test_remove(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'test')

        self.reddit.read_only = False
        with self.recorder.use_cassette('TestWikiPageModeration.test_remove'):
            page.mod.remove('bboe')

    def test_settings(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'test')

        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestWikiPageModeration.test_settings'):
            settings = page.mod.settings()
        assert {'editors': [], 'listed': True, 'permlevel': 0} == settings

    def test_update(self):
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, 'test')

        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestWikiPageModeration.test_update'):
            updated = page.mod.update(listed=False, permlevel=1)
        assert {'editors': [], 'listed': False, 'permlevel': 1} == updated



"""Provide the ListingGenerator class."""
from .listing import FlairListing
from ..base import PRAWBase


class ListingGenerator(PRAWBase):
    """Instances of this class generate ``RedditBase`` instances."""

    def __init__(self, reddit, url, limit=100, params=None):
        """Initialize a ListingGenerator instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param url: A URL returning a reddit listing.
        :param limit: The number of content entries to fetch. If ``limit`` is
            None, then fetch as many entries as possible. Most of reddit's
            listings contain a maximum of 1000 items, and are returned 100 at a
            time. This class will automatically issue all necessary
            requests. (Default: 100)
        :param params: A dictionary containing additional query string
            parameters to send with the request.

        """
        super(ListingGenerator, self).__init__(reddit, None)
        self._exhausted = False
        self._listing = None
        self._list_index = None
        self.limit = limit
        self.params = params or {}
        self.params['limit'] = limit or 1024
        self.url = url
        self.yielded = 0

    def __iter__(self):
        """Permit ListingGenerator to operate as an iterator."""
        return self

    def __next__(self):
        """Permit ListingGenerator to operate as a generator in py3."""
        if self.limit is not None and self.yielded >= self.limit:
            raise StopIteration()

        if self._listing is None or self._list_index >= len(self._listing):
            self._next_batch()

        self._list_index += 1
        self.yielded += 1
        return self._listing[self._list_index - 1]

    def _next_batch(self):
        if self._exhausted:
            raise StopIteration()

        self._listing = self._reddit.get(self.url, params=self.params)
        if isinstance(self._listing, list):
            self._listing = self._listing[1]  # for submission duplicates
        elif isinstance(self._listing, dict):
            self._listing = FlairListing(self._reddit, self._listing)
        self._list_index = 0

        if len(self._listing) == 0:
            raise StopIteration()

        if self._listing.after:
            self.params['after'] = self._listing.after
        else:
            self._exhausted = True

    def next(self):
        """Permit ListingGenerator to operate as a generator in py2."""
        return self.__next__()



"""Test praw.models.auth."""
from prawcore import InvalidToken
import pytest

from .. import UntrustedIntegrationTest


class TestAuthImplicit(UntrustedIntegrationTest):
    def test_implicit__with_invalid_token(self):
        self.reddit.auth.implicit('invalid_token', 10, 'read')
        with self.recorder.use_cassette(
                'TestAuthImplicit.test_implicit__with_invalid_token'):
            with pytest.raises(InvalidToken):
                self.reddit.user.me()



"""Provide the ListingMixin class."""
from six.moves.urllib.parse import urljoin  # pylint: disable=import-error

from ...base import PRAWBase
from ..generator import ListingGenerator


class GildedListingMixin(PRAWBase):
    """Mixes in the gilded method."""

    def gilded(self, **generator_kwargs):
        """Return a ListingGenerator for gilded items.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, urljoin(self._path, 'gilded'),
                                **generator_kwargs)



"""Provide the Front class."""
from ..const import API_PATH
from .listing.generator import ListingGenerator
from .base import PRAWBase


class Inbox(PRAWBase):
    """Inbox is a Listing class that represents the Inbox."""

    def all(self, **generator_kwargs):
        """Return a ListingGenerator for all inbox comments and messages.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, API_PATH['inbox'],
                                **generator_kwargs)

    def comment_replies(self, **generator_kwargs):
        """Return a ListingGenerator for comment replies.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, API_PATH['comment_replies'],
                                **generator_kwargs)

    def mark_read(self, items):
        """Mark Comments or Messages as read.

        :param items: A list containing Comments and Messages to be be marked
            as read relative to the authenticated user's inbox.

        Requests are batched at 25 items (reddit limit).

        """
        while items:
            data = {'id': ','.join(x.fullname for x in items[:25])}
            self._reddit.post(API_PATH['read_message'], data=data)
            items = items[25:]

    def mark_unread(self, items):
        """Unmark Comments or Messages as read.

        :param items: A list containing Comments and Messages to be be marked
            as unread relative to the authenticated user's inbox.

        Requests are batched at 25 items (reddit limit).

        """
        while items:
            data = {'id': ','.join(x.fullname for x in items[:25])}
            self._reddit.post(API_PATH['unread_message'], data=data)
            items = items[25:]

    def messages(self, **generator_kwargs):
        """Return a ListingGenerator for inbox messages.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, API_PATH['messages'],
                                **generator_kwargs)

    def submission_replies(self, **generator_kwargs):
        """Return a ListingGenerator for submission replies.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, API_PATH['submission_replies'],
                                **generator_kwargs)

    def sent(self, **generator_kwargs):
        """Return a ListingGenerator for sent messages.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, API_PATH['sent'],
                                **generator_kwargs)

    def unread(self, mark_read=False, **generator_kwargs):
        """Return a ListingGenerator for unread comments and messages.

        :param mark_read: Marks the messages as read when they're obtained
            (Default: False).

        Note: When marking messages as read, the entire batch (up to 100 at a
        time) is marked as read when fetched. Failure to consume the entire
        listing may result in missed messages if you only obtain unread
        messages.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        params = {'mark': mark_read}
        return ListingGenerator(self._reddit, API_PATH['unread'],
                                params=params, **generator_kwargs)



from praw.models import Subreddit
import mock
import pytest

from ... import IntegrationTest


class TestMultireddit(IntegrationTest):
    @mock.patch('time.sleep', return_value=None)
    def test_add(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestMultireddit.test_add'):
            multi = self.reddit.user.multireddits()[0]
            multi.add('redditdev')
            assert 'redditdev' in multi.subreddits

    @mock.patch('time.sleep', return_value=None)
    def test_copy(self, _):
        self.reddit.read_only = False
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette('TestMultireddit.test_copy'):
            new = multi.copy()
        assert new.name == multi.name
        assert new.display_name == multi.display_name
        assert pytest.placeholders.username in new.path

    @mock.patch('time.sleep', return_value=None)
    def test_copy__with_display_name(self, _):
        self.reddit.read_only = False
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        name = 'A--B\n' * 10
        with self.recorder.use_cassette(
                'TestMultireddit.test_copy__with_display_name'):
            new = multi.copy(display_name=name)
        assert new.name == 'a_b_a_b_a_b_a_b_a_b'
        assert new.display_name == name
        assert pytest.placeholders.username in new.path

    @mock.patch('time.sleep', return_value=None)
    def test_create(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestMultireddit.test_create'):
            multireddit = self.reddit.multireddit.create(
                'PRAW create test', subreddits=['redditdev'])
        assert multireddit.display_name == 'PRAW create test'
        assert multireddit.name == 'praw_create_test'

    @mock.patch('time.sleep', return_value=None)
    def test_delete(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestMultireddit.test_delete'):
            multi = self.reddit.user.multireddits()[0]
            multi.delete()

    @mock.patch('time.sleep', return_value=None)
    def test_remove(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestMultireddit.test_remove'):
            multi = self.reddit.user.multireddits()[0]
            multi.remove('redditdev')
            assert 'redditdev' not in multi.subreddits

    @mock.patch('time.sleep', return_value=None)
    def test_rename(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestMultireddit.test_rename'):
            multi = self.reddit.user.multireddits()[0]
            multi.rename('PRAW Renamed')
            assert multi.display_name == 'PRAW Renamed'
            assert multi.name == 'praw_renamed'

    def test_subreddits(self):
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette('TestMultireddit.test_subreddits'):
            assert multi.subreddits
        assert all(isinstance(x, Subreddit) for x in multi.subreddits)

    @mock.patch('time.sleep', return_value=None)
    def test_update(self, _):
        self.reddit.read_only = False
        subreddits = ['pokemongo', 'pokemongodev']
        with self.recorder.use_cassette('TestMultireddit.test_update'):
            multi = self.reddit.user.multireddits()[0]
            prev_path = multi.path
            multi.update(display_name='Updated display name',
                         subreddits=subreddits)
        assert multi.display_name == 'Updated display name'
        assert multi.path == prev_path
        assert multi.subreddits == subreddits


class TestMultiredditListings(IntegrationTest):
    def test_comments(self):
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette(
                'TestMultiredditListings.test_comments'):
            comments = list(multi.comments())
        assert len(comments) == 100

    def test_controversial(self):
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette(
                'TestMultiredditListings.test_controversial'):
            submissions = list(multi.controversial())
        assert len(submissions) == 100

    def test_gilded(self):
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette('TestMultiredditListings.test_gilded'):
            submissions = list(multi.gilded())
        assert len(submissions) == 100

    def test_hot(self):
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette('TestMultiredditListings.test_hot'):
            submissions = list(multi.hot())
        assert len(submissions) == 100

    def test_new(self):
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette('TestMultiredditListings.test_new'):
            submissions = list(multi.new())
        assert len(submissions) == 100

    @mock.patch('time.sleep', return_value=None)
    def test_new__self_multi(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestMultiredditListings.test_new__self_multi'):
            multi = self.reddit.user.multireddits()[0]
            submissions = list(multi.new())
        assert len(submissions) == 100

    def test_rising(self):
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette('TestMultiredditListings.test_rising'):
            submissions = list(multi.rising())
        assert len(submissions) > 0

    def test_top(self):
        multi = self.reddit.multireddit('kjoneslol', 'sfwpornnetwork')
        with self.recorder.use_cassette('TestMultiredditListings.test_top'):
            submissions = list(multi.top())
        assert len(submissions) == 100



"""Provide the Comment class."""
from six.moves.urllib.parse import urljoin  # pylint: disable=import-error

from .base import RedditBase
from .mixins import InboxableMixin, UserContentMixin
from .redditor import Redditor


class Comment(RedditBase, InboxableMixin, UserContentMixin):
    """A class that represents a reddit comments."""

    STR_FIELD = 'id'

    @property
    def is_root(self):
        """Return True when the comment is a top level comment."""
        parent_type = self.parent_id.split('_', 1)[0]
        return parent_type == self._reddit.config.kinds['submission']

    @property
    def submission(self):
        """Return the Submission object this comment belongs to."""
        if not self._submission:  # Comment not from submission
            self._submission = self._reddit.submission(
                self.link_id.split('_', 1)[1])
        return self._submission

    @submission.setter
    def submission(self, submission):
        """Update the Submission associated with the Comment."""
        assert self.name not in submission.comments._comments_by_id
        submission.comments._comments_by_id[self.name] = self
        self._submission = submission
        for reply in getattr(self, 'replies', []):
            reply.submission = submission

    def __init__(self, reddit, id=None,  # pylint: disable=redefined-builtin
                 _data=None):
        """Construct an instance of the Comment object."""
        if bool(id) == bool(_data):
            raise TypeError('Either `id` or `_data` must be provided.')
        super(Comment, self).__init__(reddit, _data)
        self._submission = None
        if id:
            self.id = id  # pylint: disable=invalid-name

    def __setattr__(self, attribute, value):
        """Objectify author, replies, and subreddit."""
        # pylint: disable=redefined-variable-type
        if attribute == 'author':
            value = Redditor.from_data(self._reddit, value)
        elif attribute == 'replies':
            # TODO: From info URL replies = "", even if there are replies.
            # From a comment tree replies = "" when it has no replies.
            if value == '':
                value = []
            else:
                value = self._reddit._objector.objectify(value).children
        elif attribute == 'subreddit':
            value = self._reddit.subreddit(value)
        super(Comment, self).__setattr__(attribute, value)

    def permalink(self, fast=False):
        """Return a permalink to the comment.

        :param fast: Return the result as quickly as possible (Default: False).

        In order to determine the full permalink for a comment, the Submission
        may need to be fetched if it hasn't been already. Set ``fast=True`` if
        you want to bypass that possible load.

        A full permalink looks like:
        /r/redditdev/comments/2gmzqe/praw_https_enabled/cklhv0f

        A fast-loaded permalink for the same comment will look like:
        /comments/2gmzqe//cklhv0f

        """
        # pylint: disable=no-member
        if not fast or 'permalink' in self.submission.__dict__:
            return urljoin(self.submission.permalink, self.id)
        return '/comments/{}//{}'.format(self.submission.id, self.id)



"""Provide the VotableMixin class."""
from ....const import API_PATH


class VotableMixin(object):
    """Interface for RedditBase classes that can be voted on."""

    def _vote(self, direction):
        self._reddit.post(API_PATH['vote'], data={'dir': str(direction),
                                                  'id': self.fullname})

    def clear_vote(self):
        """Clear the authenticated user's vote on the object.

        .. note:: votes must be cast by humans. That is, API clients proxying a
           human's action one-for-one are OK, but bots deciding how to vote on
           content or amplifying a human's vote are not. See the reddit rules
           for more details on what constitutes vote cheating. [`Ref
           <https://www.reddit.com/dev/api#POST_api_vote>`_]

        """
        self._vote(direction=0)

    def downvote(self):
        """Downvote the object.

        .. note:: votes must be cast by humans. That is, API clients proxying a
           human's action one-for-one are OK, but bots deciding how to vote on
           content or amplifying a human's vote are not. See the reddit rules
           for more details on what constitutes vote cheating. [`Ref
           <https://www.reddit.com/dev/api#POST_api_vote>`_]

        """
        self._vote(direction=-1)

    def upvote(self):
        """Upvote the object.

        .. note:: votes must be cast by humans. That is, API clients proxying a
           human's action one-for-one are OK, but bots deciding how to vote on
           content or amplifying a human's vote are not. See the reddit rules
           for more details on what constitutes vote cheating. [`Ref
           <https://www.reddit.com/dev/api#POST_api_vote>`_]

        """
        self._vote(direction=1)



"""Provide the BaseList class."""
from ..base import PRAWBase


class BaseList(PRAWBase):
    """An abstract class to coerce a list into a PRAWBase."""

    CHILD_ATTRIBUTE = None

    @staticmethod
    def _convert(reddit, item):
        raise NotImplementedError('BaseList must be extended.')

    def __init__(self, reddit, _data):
        """Initialize a BaseList instance.

        :param reddit: An instance of :class:`~.Reddit`.

        """
        super(BaseList, self).__init__(reddit, _data)

        if self.CHILD_ATTRIBUTE is None:
            raise NotImplementedError('BaseList must be extended.')

        child_list = getattr(self, self.CHILD_ATTRIBUTE)
        for index, item in enumerate(child_list):
            child_list[index] = self._convert(reddit, item)

    def __contains__(self, item):
        """Test if item exists in the list."""
        return item in getattr(self, self.CHILD_ATTRIBUTE)

    def __getitem__(self, index):
        """Return the item at position index in the list."""
        return getattr(self, self.CHILD_ATTRIBUTE)[index]

    def __iter__(self):
        """Return an iterator to the list."""
        return getattr(self, self.CHILD_ATTRIBUTE).__iter__()

    def __len__(self):
        """Return the number of items in the list."""
        return len(getattr(self, self.CHILD_ATTRIBUTE))

    def __str__(self):
        """Return a string representation of the list."""
        return str(getattr(self, self.CHILD_ATTRIBUTE))



"""Provide the MoreComments class."""
from ...const import API_PATH
from ..base import PRAWBase


class MoreComments(PRAWBase):
    """A class indicating there are more comments."""

    def __init__(self, reddit, _data):
        """Construct an instance of the MoreComments object."""
        self.count = self.parent_id = None
        self.children = []
        super(MoreComments, self).__init__(reddit, _data)
        self._comments = None
        self.submission = None

    def __eq__(self, other):
        """Return True if these MoreComments instances are the same."""
        return isinstance(other, self.__class__) and \
            self.count == other.count and self.children == other.children

    def __lt__(self, other):
        """Proide a sort order on the MoreComments object."""
        # To work with heapq a "smaller" item is the one with the most comments
        # We are intentionally making the biggest element the smallest element
        # to turn the min-heap implementation in heapq into a max-heap.
        return self.count > other.count

    def __repr__(self):
        """Return a string representation of the MoreComments instance."""
        children = self.children[:4]
        if len(self.children) > 4:
            children[-1] = '...'
        return '<{} count={}, children={!r}>'.format(
            self.__class__.__name__, self.count, children)

    def _continue_comments(self, update):
        assert len(self.children) == 0
        parent = self._load_comment(self.parent_id.split('_', 1)[1])
        self._comments = parent.replies
        if update:
            for comment in self._comments:
                comment.submission = self.submission
        return self._comments

    def _load_comment(self, comment_id):
        path = '{}_/{}'.format(
            API_PATH['submission'].format(id=self.submission.id), comment_id)
        _, comments = self._reddit.get(path, params={
            'limit': self.submission.comment_limit,
            'sort': self.submission.comment_sort})
        assert len(comments.children) == 1
        return comments.children[0]

    def comments(self, update=True):
        """Fetch and return the comments for a single MoreComments object."""
        if self._comments is None:
            if self.count == 0:  # Handle 'continue this thread'
                return self._continue_comments(update)
            assert self.children
            data = {'children': ','.join(self.children),
                    'link_id': self.submission.fullname,
                    'sort': self.submission.comment_sort}
            self._comments = self._reddit.post(API_PATH['morechildren'],
                                               data=data)
            if update:
                for comment in self._comments:
                    comment.submission = self.submission
        return self._comments



"""Test praw.reddit."""
import mock

from praw.models import LiveThread

from . import IntegrationTest


class TestReddit(IntegrationTest):
    @mock.patch('time.sleep', return_value=None)
    def test_live_create(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestReddit.test_live_create'):
            live = self.reddit.live.create('PRAW Create Test')
            assert isinstance(live, LiveThread)
            assert live.title == 'PRAW Create Test'

    def test_random_subreddit(self):
        names = set()
        with self.recorder.use_cassette(
                'TestReddit.test_random_subreddit'):
            for i in range(3):
                names.add(self.reddit.random_subreddit().display_name)
        assert len(names) == 3

    def test_subreddit_with_randnsfw(self):
        with self.recorder.use_cassette(
                'TestReddit.test_subreddit_with_randnsfw'):
            subreddit = self.reddit.subreddit('randnsfw')
            assert subreddit.display_name != 'randnsfw'
            assert subreddit.over18

    def test_subreddit_with_random(self):
        with self.recorder.use_cassette(
                'TestReddit.test_subreddit_with_random'):
            assert self.reddit.subreddit('random').display_name != 'random'



from praw.models import Message, SubredditMessage
import mock
import pytest

from ... import IntegrationTest


class TestMessage(IntegrationTest):
    @mock.patch('time.sleep', return_value=None)
    def test_block(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestMessage.test_block'):
            message = None
            for item in self.reddit.inbox.messages():
                if item.author and item.author != pytest.placeholders.username:
                    message = item
                    break
            else:
                assert False, 'no message found'
            message.block()

    @mock.patch('time.sleep', return_value=None)
    def test_mark_read(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestMessage.test_mark_read'):
            message = None
            for item in self.reddit.inbox.unread():
                if isinstance(item, Message):
                    message = item
                    break
            else:
                assert False, 'no message found in unread'
            message.mark_read()

    @mock.patch('time.sleep', return_value=None)
    def test_mark_unread(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestMessage.test_mark_unread'):
            message = next(self.reddit.inbox.messages())
            message.mark_unread()


class TestSubredditMessage(IntegrationTest):
    def test_mute(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditMessage.test_mute'):
            message = SubredditMessage(self.reddit, _data={'id': '5yr8id'})
            message.mute()

    def test_unmute(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditMessage.test_unmute'):
            message = SubredditMessage(self.reddit, _data={'id': '5yr8id'})
            message.unmute()



"""Provide the MessageableMixin class."""
from ....const import API_PATH


class MessageableMixin(object):
    """Interface for classes that can be messaged."""

    def message(self, subject, message, from_subreddit=None):
        """
        Send a message to a redditor or a subreddit's moderators (mod mail).

        :param subject: The subject of the message.
        :param message: The message content.
        :param from_subreddit: A Subreddit instance or string to send the
            message from. When provided, messages are sent from the subreddit
            rather than from the authenticated user. Note that the
            authenticated user must be a moderator of the subreddit and have
            mail permissions.
        :returns: The json response from the server.

        """
        data = {'subject': subject, 'text': message,
                'to': '{}{}'.format(getattr(
                    self.__class__, 'MESSAGE_PREFIX', ''), self)}
        if from_subreddit:
            data['from_sr'] = str(from_subreddit)
        return self._reddit.post(API_PATH['compose'], data=data)



"""Provide the helper classes."""
from json import dumps

from ..const import API_PATH
from .base import PRAWBase
from .reddit.multi import Multireddit, Subreddit


class LiveHelper(PRAWBase):
    """Provide a set of functions to interact with LiveThreads."""

    def create(self, title, description=None, nsfw=False, resources=None):
        """Create a new LiveThread.

        :param title: The title of the new LiveThread.
        :param description: (Optional) The new LiveThread's description.
        :param nsfw: (boolean) Indicate whether this thread is not safe for
            work (default: False).
        :param resources: (Optional) Markdown formatted information that is
            useful for the LiveThread.
        :returns: The new LiveThread object.

        """
        return self._reddit.post(API_PATH['livecreate'], data={
            'description': description, 'nsfw': nsfw, 'resources': resources,
            'title': title})


class MultiredditHelper(PRAWBase):
    """Provide a set of functions to interact with Multireddits."""

    def __call__(self, redditor, name):
        """Return a lazy instance of :class:`~.Multireddit`.

        :param redditor: A string or :class:`~.Redditor` instance who owns the
            multireddit.
        :param name: The name of the multireddit.

        """
        path = '/user/{}/m/{}'.format(redditor, name)
        return Multireddit(self._reddit, _data={'name': name, 'path': path})

    def create(self, display_name, subreddits, description_md=None,
               icon_name=None, key_color=None, visibility='private',
               weighting_scheme='classic'):
        """Create a new multireddit.

        :param display_name: The display name for the new multireddit.
        :param subreddits: Subreddits to add to the new multireddit.
        :param description_md: (Optional) Description for the new multireddit,
            formatted in markdown.
        :param icon_name: (Optional) Can be one of: ``art
            and design``, ``ask``, ``books``, ``business``, ``cars``,
            ``comics``, ``cute animals``, ``diy``, ``entertainment``, ``food
            and drink``, ``funny``, ``games``, ``grooming``, ``health``, ``life
            advice``, ``military``, ``models pinup``, ``music``, ``news``,
            ``philosophy``, ``pictures and gifs``, ``science``, ``shopping``,
            ``sports``, ``style``, ``tech``, ``travel``, ``unusual stories``,
            ``video``, or ``None``.
        :param key_color: (Optional) RGB hex color code of the form `#FFFFFF`.
        :param visibility: (Optional) Can be one of: ``hidden``, ``private``,
            ``public`` (Default: private).
        :param weighting_scheme: (Optional) Can be one of: ``classic``,
            ``fresh`` (Default: classic)
        :returns: The new Multireddit object.

        """
        model = {'description_md': description_md,
                 'display_name': display_name,
                 'icon_name': icon_name,
                 'key_color': key_color,
                 'subreddits': [{'name': str(sub)} for sub in subreddits],
                 'visibility': visibility,
                 'weighting_scheme': weighting_scheme}
        return self._reddit.post(API_PATH['multireddit_base'],
                                 data={'model': dumps(model)})


class SubredditHelper(PRAWBase):
    """Provide a set of functions to interact with Subreddits."""

    def __call__(self, display_name):
        """Return a lazy instance of :class:`~.Subreddit`.

        :param display_name: The name of the subreddit.
        """
        lower_name = display_name.lower()

        if lower_name == 'random':
            return self._reddit.random_subreddit()
        elif lower_name == 'randnsfw':
            return self._reddit.random_subreddit(nsfw=True)

        return Subreddit(self._reddit, display_name=display_name)

    def create(self, name, link_type, subreddit_type, title, wikimode,
               **other_settings):
        """
        Create a new subreddit.

        :param link_type: The types of submissions users can make.
            One of ``any``, ``link``, ``self``.
        :param name: The name for the new subreddit.
        :param subreddit_type: One of ``archived``, ``employees_only``,
            ``gold_only``, ``gold_restricted``, ``private``, ``public``,
            ``restricted``.
        :param title: The title of the subreddit.
        :param wikimode: One of  ``anyone``, ``disabled``, ``modonly``.

        See :meth:`~.SubredditModeration.update` for documentation of other
        available settings.

        Any keyword parameters not provided, or set explicitly to None, will
        take on a default value assigned by the Reddit server.

        """
        Subreddit._create_or_update(_reddit=self._reddit, name=name,
                                    link_type=link_type,
                                    subreddit_type=subreddit_type,
                                    title=title, wikimode=wikimode,
                                    **other_settings)
        return self(name)



"""Provide the RedditBase class."""
from ...const import API_PATH
from ..base import PRAWBase


class RedditBase(PRAWBase):
    """Base class that represents actual Reddit objects."""

    @property
    def fullname(self):
        """Return the object's fullname.

        A fullname is an object's kind mapping like `t3` followed by an
        underscore and the object's base36 ID, e.g., `t1_c5s96e0`.

        """
        return '{}_{}'.format(self._reddit._objector.kind(self),
                              self.id)  # pylint: disable=invalid-name

    def __eq__(self, other):
        """Return whether the other instance equals the current."""
        if isinstance(other, str):
            return other.lower() == str(self).lower()
        return (isinstance(other, self.__class__) and
                str(self).lower() == str(other).lower())

    def __getattr__(self, attribute):
        """Return the value of `attrribute`."""
        if not attribute.startswith('__') and not self._fetched:
            self._fetch()
            return getattr(self, attribute)
        raise AttributeError('{!r} object has no attribute {!r}'
                             .format(self.__class__.__name__, attribute))

    def __hash__(self):
        """Return the hash of the current instance."""
        return hash(self.__class__.__name__) ^ hash(str(self).lower())

    def __init__(self, reddit, _data):
        """Initialize a RedditBase instance.

        :param reddit: An instance of :class:`~.Reddit`.

        """
        super(RedditBase, self).__init__(reddit, _data)
        self._fetched = False

    def __repr__(self):
        """Return an object initialization representation of the instance."""
        return '{}({}={!r})'.format(self.__class__.__name__, self.STR_FIELD,
                                    str(self))

    def __str__(self):
        """Return a string representation of the instance."""
        return getattr(self, self.STR_FIELD)

    def __ne__(self, other):
        """Return whether the other instance differs from the current."""
        return not self == other

    def _fetch(self):
        if '_info_path' in dir(self):
            other = self._reddit.get(self._info_path())
        else:
            other = self._reddit.get(API_PATH['info'],
                                     params={'id': self.fullname}).children[0]
        self.__dict__.update(other.__dict__)
        self._fetched = True

    def _reset_attributes(self, *attributes):
        for attribute in attributes:
            del self.__dict__[attribute]
        self._fetched = False



"""Provide the Subreddit class."""
from prawcore import Redirect
from six.moves.urllib.parse import urljoin  # pylint: disable=import-error

from ...const import API_PATH
from ..util import stream_generator
from ..listing.generator import ListingGenerator
from ..listing.mixins import SubredditListingMixin
from .base import RedditBase
from .mixins import MessageableMixin
from .wikipage import WikiPage


class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
    """A class for Subreddits."""

    STR_FIELD = 'display_name'
    MESSAGE_PREFIX = '#'

    @staticmethod
    def _create_or_update(_reddit, allow_images=None, allow_top=None,
                          collapse_deleted_comments=None,
                          comment_score_hide_mins=None, description=None,
                          domain=None, exclude_modqueue_banned=None,
                          header_hover_text=None, hide_ads=None, lang=None,
                          key_color=None, link_type=None, name=None,
                          over_18=None, public_description=None,
                          public_traffic=None, show_media=None,
                          show_thumbnails=None, spam_comments=None,
                          # pylint: disable=invalid-name, too-many-locals
                          spam_links=None, spam_selfposts=None, sr=None,
                          submit_link_label=None, submit_text=None,
                          submit_text_label=None, subreddit_type=None,
                          suggested_comment_sort=None, title=None,
                          wiki_edit_age=None, wiki_edit_karma=None,
                          wikimode=None, **other_settings):
        model = {'allow_images': allow_images,
                 'allow_top': allow_top,
                 'collapse_deleted_comments': collapse_deleted_comments,
                 'comment_score_hide_mins': comment_score_hide_mins,
                 'description': description,
                 'domain': domain,
                 'exclude_modqueue_banned': exclude_modqueue_banned,
                 'header_hover_text': header_hover_text,
                 'hide_ads': hide_ads,
                 'key_color': key_color,
                 'lang': lang,
                 'link_type': link_type,
                 'name': name,
                 'over_18': over_18,
                 'public_description': public_description,
                 'public_traffic': public_traffic,
                 'show_media': show_thumbnails,
                 'show_media_preview': show_media,
                 'spam_comments': spam_comments,
                 'spam_links': spam_links,
                 'spam_selfposts': spam_selfposts,
                 'sr': sr,
                 'submit_link_label': submit_link_label,
                 'submit_text': submit_text,
                 'submit_text_label': submit_text_label,
                 'suggested_comment_sort': suggested_comment_sort,
                 'title': title,
                 'type': subreddit_type,
                 'wiki_edit_age': wiki_edit_age,
                 'wiki_edit_karma': wiki_edit_karma,
                 'wikimode': wikimode}

        model.update(other_settings)

        _reddit.post(API_PATH['site_admin'], data=model)

    @property
    def flair(self):
        """An instance of :class:`.SubredditFlair`."""
        if self._flair is None:
            self._flair = SubredditFlair(self)
        return self._flair

    @property
    def mod(self):
        """An instance of :class:`.SubredditModeration`."""
        if self._mod is None:
            self._mod = SubredditModeration(self)
        return self._mod

    @property
    def stream(self):
        """An instance of :class:`.SubredditStream`."""
        if self._stream is None:
            self._stream = SubredditStream(self)
        return self._stream

    @property
    def wiki(self):
        """An instance of :class:`.SubredditWiki`."""
        if self._wiki is None:
            self._wiki = SubredditWiki(self)
        return self._wiki

    def __init__(self, reddit, display_name=None, _data=None):
        """Initialize a Subreddit instance.

        :param reddit: An instance of :class:`~.Reddit`.
        :param display_name: The name of the subreddit.

        """
        if bool(display_name) == bool(_data):
            raise TypeError(
                'Either `display_name` or `_data` must be provided.')
        super(Subreddit, self).__init__(reddit, _data)
        if display_name:
            self.display_name = display_name
        self._flair = self._mod = self._stream = self._wiki = None
        self._path = API_PATH['subreddit'].format(subreddit=self)
        self._prepare_relationships()

    def _info_path(self):
        return API_PATH['subreddit_about'].format(subreddit=self)

    def _prepare_relationships(self):
        for relationship in ['banned', 'contributor', 'moderator', 'muted']:
            setattr(self, relationship,
                    SubredditRelationship(self, relationship))

    def random(self):
        """Return a random Submission."""
        url = API_PATH['subreddit_random'].format(subreddit=self)
        try:
            self._reddit.get(url, params={'unique': self._reddit._next_unique})
        except Redirect as redirect:
            path = redirect.path
        return self._submission_class(self._reddit, url=urljoin(
            self._reddit.config.reddit_url, path))

    def search(self, query, sort='relevance', syntax='cloudsearch',
               time_filter='all', **generator_kwargs):
        """Return a ListingGenerator for items that match ``query``.

        :param query: The query string to search for.
        :param sort: Can be one of: relevance, hot, top, new,
            comments. (default: relevance).
        :param syntax: Can be one of: cloudsearch, lucene, plain
            (default: cloudsearch).
        :param time_filter: Can be one of: all, day, hour, month, week, year
            (default: all).

        For more information on building a search query see:
            https://www.reddit.com/wiki/search

        """
        self.validate_time_filter(time_filter)
        params = generator_kwargs.setdefault('params', {})
        params.update({'q': query, 'restrict_sr': True, 'sort': sort,
                       'syntax': syntax, 't': time_filter})
        url = API_PATH['search'].format(subreddit=self)
        return ListingGenerator(self._reddit, url, **generator_kwargs)

    def submit(self, title, selftext=None, url=None, resubmit=True,
               send_replies=True):
        """Add a submission to the subreddit.

        :param title: The title of the submission.
        :param selftext: The markdown formatted content for a ``text``
            submission.
        :param url: The URL for a ``link`` submission.
        :param resubmit: When False, an error will occur if the URL has already
            been submitted (Default: True).
        :param send_replies: When True, messages will be sent to the submission
            author when comments are made to the submission (Default: True).
        :returns: A :class:`~.Submission` object for the newly created
            submission.

        Either ``selftext`` or ``url`` can be provided, but not both.

        """
        if bool(selftext) == bool(url):
            raise TypeError('Either `selftext` or `url` must be provided.')

        data = {'sr': str(self), 'resubmit': bool(resubmit),
                'sendreplies': bool(send_replies), 'title': title}
        if selftext is not None:
            data.update(kind='self', text=selftext)
        else:
            data.update(kind='link', url=url)
        return self._reddit.post(API_PATH['submit'], data=data)


class SubredditFlair(object):
    """Provide a set of functions to interact with a Subreddit's flair."""

    def __init__(self, subreddit):
        """Create a SubredditFlair instance.

        :param subreddit: The subreddit whose flair to work with.

        """
        self.subreddit = subreddit

    def __iter__(self):
        """Iterate through the Redditors and their associated flair."""
        url = API_PATH['flairlist'].format(subreddit=self.subreddit)
        params = {'unique': self.subreddit._reddit._next_unique}
        for item in ListingGenerator(self.subreddit._reddit, url, None,
                                     params=params):
            yield item

    def delete_all(self):
        """Delete all Redditor flair in the Subreddit.

        :returns: List of dictionaries indicating the success or failure of
            each delete.

        """
        return self.update(x['user'] for x in self)

    def set(self, thing, text='', css_class=''):
        """Set flair for a Redditor or Submission.

        :param thing: An instance of Redditor or Submission, or a string. When
            a string is provided it will be treated as the name of a Redditor.
        :param text: The flair text to associate with the Redditor or
            Submission (Default: '').
        :param css_class: The css class to associate with the flair html
            (Default: '').

        This method can only be used by an authenticated user who is a
        moderator of the associated Subreddit.

        """
        data = {'css_class': css_class, 'text': text}
        if thing.__class__.__name__ == 'Submission':
            data['link'] = thing.fullname
        else:
            data['name'] = str(thing)
        url = API_PATH['flair'].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url, data=data)

    def update(self, flair_list, text='', css_class=''):
        """Set or clear the flair for many Redditors at once.

        :param redditor_flair_list: Each item in this list should be either: a
            Redditor name string, a Redditor, or a dictionary containing the
            keys ``user``, ``flair_text``, and ``flair_css_class``. The
            ``user`` key should map to a Redditor name string, or a
            Redditor. When a dictionary isn't provided, or the dictionary is
            missing one of ``flair_text``, or ``flair_css_class`` attributes
            the default values will come from the the following arguments.
        :param text: The flair text to use when not explicitly provided in
            ``flair_list`` (Default: '').
        :param css_class: The css class to use when not explicitly provided in
            ``flair_list`` (Default: '').
        :returns: List of dictionaries indicating the success or failure of
            each update.

        """
        lines = []
        for item in flair_list:
            if isinstance(item, dict):
                fmt_data = (str(item['user']), item.get('flair_text', text),
                            item.get('flair_css_class', css_class))
            else:
                fmt_data = (str(item), text, css_class)
            lines.append('"{}","{}","{}"'.format(*fmt_data))

        response = []
        url = API_PATH['flaircsv'].format(subreddit=self.subreddit)
        while len(lines):
            data = {'flair_csv': '\n'.join(lines[:100])}
            response.extend(self.subreddit._reddit.post(url, data=data))
            lines = lines[100:]
        return response


class SubredditModeration(object):
    """Provides a set of moderation functions to a Subreddit."""

    def __init__(self, subreddit):
        """Create a SubredditModeration instance.

        :param subreddit: The subreddit to moderate.

        """
        self.subreddit = subreddit

    def approve(self, thing):
        """Approve a Comment or Submission.

        :param thing: An instance of Comment or Submission.

        Approving a comment or submission reverts a removal, resets the report
        counter, adds a green check mark indicator (only visible to other
        moderators) on the website view, and sets the ``approved_by`` attribute
        to the authenticated user.

        """
        self.subreddit._reddit.post(API_PATH['approve'],
                                    data={'id': thing.fullname})

    def distinguish(self, thing, how='yes'):
        """Distinguish a Comment or Submission.

        :param thing: An instance of Comment or Submission.

        :param how: One of 'yes', 'no', 'admin', 'special'. 'yes' adds a
            moderator level distinguish. 'no' removes any distinction. 'admin'
            and 'special' require special user priviliges to use.

        """
        return self.subreddit._reddit.post(
            API_PATH['distinguish'], data={'how': how, 'id': thing.fullname})

    def ignore_reports(self, thing):
        """Ignore future reports on a Comment or Submission.

        :param thing: An instance of Comment or Submission.

        Calling this method will prevent future reports on this Comment or
        Submission from both triggering notifications and appearing in the
        various moderation listings. The report count will still increment on
        the Comment or Submission.

        """
        self.subreddit._reddit.post(API_PATH['ignore_reports'],
                                    data={'id': thing.fullname})

    def inbox(self, **generator_kwargs):
        """Return a ListingGenerator for moderator messages.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        See ``unread`` for unread moderator messages.

        """
        return ListingGenerator(
            self.subreddit._reddit, API_PATH['moderator_messages'].format(
                subreddit=self.subreddit), **generator_kwargs)

    def remove(self, thing, spam=False):
        """Remove a Comment or Submission.

        :param thing: An instance of Comment or Submission.
        :param spam: When True, use the removal to help train the Subreddit's
            spam filter (Default: False)

        """
        data = {'id': thing.fullname, 'spam': bool(spam)}
        self.subreddit._reddit.post(API_PATH['remove'], data=data)

    def settings(self):
        """Return a dictionary of the subreddit's current settings."""
        url = API_PATH['subreddit_settings'].format(subreddit=self.subreddit)
        return self.subreddit._reddit.get(url)['data']

    def undistinguish(self, thing):
        """Remove mod, admin or special distinguishing on object.

        :returns: The json response from the server.

        """
        return self.distinguish(thing, how='no')

    def unignore_reports(self, thing):
        """Resume receiving future reports on a Comment or Submission.

        :param thing: An instance of Comment or Submission.

        Future reports on this Comment or Submission will cause notifications,
        and appear in the various moderation listings.

        """
        self.subreddit._reddit.post(API_PATH['unignore_reports'],
                                    data={'id': thing.fullname})

    def unread(self, **generator_kwargs):
        """Return a ListingGenerator for unread moderator messages.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        See ``inbox`` for all messages.

        """
        return ListingGenerator(
            self.subreddit._reddit, API_PATH['moderator_unread'].format(
                subreddit=self.subreddit), **generator_kwargs)

    def update(self, **settings):
        """Update the subreddit's settings.

        :param allow_images: Allow users to upload images using the native
            image hosting. Only applies to link-only subreddits.
        :param allow_top: Allow the subreddit to appear on /r/all as well as
            the default and trending lists.
        :param collapse_deleted_comments: Collapse deleted and removed comments
            on comments pages by default.
        :param comment_score_hide_mins: The number of minutes to hide comment
            scores.
        :param description: Shown in the sidebar of your subreddit.
        :param domain: Domain name with a cname that points to
            {subreddit}.reddit.com.
        :param exclude_modqueue_banned: Exclude posts by site-wide banned users
            from modqueue/unmoderated.
        :param header_hover_text: The text seen when hovering over the snoo.
        :param hide_ads: Don't show ads within this subreddit. Only applies to
            gold-user only subreddits.
        :param key_color: A 6-digit rgb hex color (e.g. `#AABBCC`), used as a
            thematic color for your subreddit on mobile.
        :param lang: A valid IETF language tag (underscore separated).
        :param link_type: The types of submissions users can make.
            One of ``any``, ``link``, ``self``.
        :param over_18: Viewers must be over 18 years old (i.e. NSFW).
        :param public_description: Public description blurb. Appears in search
            results and on the landing page for private subreddits.
        :param public_traffic: Make the traffic stats page public.
        :param show_media: Expand media previews on comments pages.
        :param show_thumbnails: Show thumbnails on submissions.
        :param spam_comments: Spam filter strength for comments.
            One of ``all``, ``low``, ``high``.
        :param spam_links: Spam filter strength for links.
            One of ``all``, ``low``, ``high``.
        :param spam_selfposts: Spam filter strength for selfposts.
            One of ``all``, ``low``, ``high``.
        :param sr: The fullname of the subreddit whose settings will be
            updated.
        :param submit_link_label: Custom label for submit link button
            (None for default).
        :param submit_text: Text to show on submission page.
        :param submit_text_label: Custom label for submit text post button
            (None for default).
        :param subreddit_type: One of ``archived``, ``employees_only``,
            ``gold_only``, ``gold_restricted``, ``private``, ``public``,
            ``restricted``.
        :param suggested_comment_sort: All comment threads will use this
            sorting method by default. Leave None, or choose one of
            ``confidence``, ``controversial``, ``new``, ``old``, ``qa``,
            ``random``, ``top``.
        :param title: The title of the subreddit.
        :param wiki_edit_age: Account age, in days, required to edit and create
            wiki pages.
        :param wiki_edit_karma: Subreddit karma required to edit and create
            wiki pages.
        :param wikimode: One of  ``anyone``, ``disabled``, ``modonly``.

        Additional keyword arguments can be provided to handle new settings as
        Reddit introduces them.

        Unspecified settings will maintain their current value.

        """
        current_settings = self.settings()
        fullname = current_settings.pop('subreddit_id')

        # These attributes come out using different names than they go in.
        remap = {'allow_top': 'default_set',
                 'exclude_modqueue_banned': 'exclude_banned_modqueue',
                 'lang': 'language',
                 'link_type': 'content_options'}
        for (new, old) in remap.items():
            current_settings[new] = current_settings.pop(old)

        current_settings.update(settings)
        return Subreddit._create_or_update(_reddit=self.subreddit._reddit,
                                           sr=fullname, **current_settings)


class SubredditRelationship(object):
    """Represents a relationship between a redditor and subreddit."""

    def __init__(self, subreddit, relationship):
        """Create a SubredditRelationship instance.

        :param subreddit: The subreddit for the relationship.
        :param relationship: The name of the relationship.

        """
        self.relationship = relationship
        self.subreddit = subreddit

    def __iter__(self):
        """Iterate through the Redditors belonging to this relationship."""
        url = API_PATH[self.relationship].format(subreddit=self.subreddit)
        params = {'unique': self.subreddit._reddit._next_unique}
        for item in self.subreddit._reddit.get(url, params=params):
            yield item

    def add(self, redditor):
        """Add ``redditor`` to this relationship.

        :param redditor: A string or :class:`~.Redditor` instance.

        """
        data = {'name': str(redditor), 'r': str(self.subreddit),
                'type': self.relationship}
        return self.subreddit._reddit.post(API_PATH['friend'], data=data)

    def remove(self, redditor):
        """Remove ``redditor`` from this relationship.

        :param redditor: A string or :class:`~.Redditor` instance.

        """
        data = {'name': str(redditor), 'r': str(self.subreddit),
                'type': self.relationship}
        return self.subreddit._reddit.post(API_PATH['unfriend'], data=data)


class SubredditStream(object):
    """Provides submission and comment streams."""

    def __init__(self, subreddit):
        """Create a SubredditStream instance.

        :param subreddit: The subreddit associated with the streams.

        """
        self.subreddit = subreddit

    def comments(self):
        """Yield new comments as they become available.

        Comments are yielded oldest first. Up to 100 historial comments will
        initially be returned.

        """
        return stream_generator(self.subreddit.comments)

    def submissions(self):
        """Yield new submissions as they become available.

        Submissions are yielded oldest first. Up to 100 historial submissions
        will initially be returned.

        """
        return stream_generator(self.subreddit.new)


class SubredditWiki(object):
    """Provides a set of moderation functions to a Subreddit."""

    def __getitem__(self, page_name):
        """Lazily return the WikiPage for the subreddit named ``page_name``."""
        return WikiPage(self.subreddit._reddit, self.subreddit, page_name)

    def __init__(self, subreddit):
        """Create a SubredditModeration instance.

        :param subreddit: The subreddit to moderate.

        """
        self.banned = SubredditRelationship(subreddit, 'wikibanned')
        self.contributors = SubredditRelationship(subreddit, 'wikicontributor')
        self.subreddit = subreddit

    def __iter__(self):
        """Iterate through the pages of the wiki."""
        response = self.subreddit._reddit.get(
            API_PATH['wiki_pages'].format(subreddit=self.subreddit),
            params={'unique': self.subreddit._reddit._next_unique})
        for page_name in response['data']:
            yield WikiPage(self.subreddit._reddit, self.subreddit, page_name)

    def create(self, name, content, reason=None, **other_settings):
        """Create a new wiki page.

        :param name: The name of the new WikiPage. This name will be normalied.
        :param content: The content of the new WikiPage.
        :param reason: (Optional) The reason for the creation.
        :param other_settings: Additional keyword arguments to pass.

        """
        name = name.replace(' ', '_').lower()
        new = WikiPage(self.subreddit._reddit, self.subreddit, name)
        new.edit(content=content, reason=reason, **other_settings)
        return new



"""Provide the EditableMixin class."""
from ....const import API_PATH


class EditableMixin(object):
    """Interface for classes that can be edited and deleted."""

    def delete(self):
        """Delete the object."""
        self._reddit.post(API_PATH['del'], {'id': self.fullname})

    def edit(self, body):
        """Replace the body of the object with `body`.

        :param body: The markdown formatted content for the updated object.
        :returns: The current instance after updating its attributes.

        """
        data = {'text': body, 'thing_id': self.fullname}
        updated = self._reddit.post(API_PATH['edit'], data=data)[0]
        for attribute in ['_fetched', '_reddit', '_submission', 'replies',
                          'subreddit']:
            if attribute in updated.__dict__:
                delattr(updated, attribute)
        self.__dict__.update(updated.__dict__)
        return self



"""Test praw.models.inbox."""
from praw.models import Comment, Message
import mock

from .. import IntegrationTest


class TestInbox(IntegrationTest):
    def test_all(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestInbox.test_all'):
            count = 0
            for item in self.reddit.inbox.all():
                assert isinstance(item, Comment) or isinstance(item, Message)
                count += 1
            assert count == 100

    @mock.patch('time.sleep', return_value=None)
    def test_all__with_limit(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestInbox.test_all__with_limit'):
            assert len(list(self.reddit.inbox.all(limit=128))) == 128

    def test_comment_replies(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestInbox.test_comment_replies'):
            count = 0
            for item in self.reddit.inbox.comment_replies(limit=64):
                assert isinstance(item, Comment)
                assert item.parent_id.startswith(
                    self.reddit.config.kinds['comment'])
                count += 1
            assert count == 64

    @mock.patch('time.sleep', return_value=None)
    def test_mark_read(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestInbox.test_mark_read',
                match_requests_on=['uri', 'method', 'body']):
            self.reddit.inbox.mark_read(list(self.reddit.inbox.unread()))

    @mock.patch('time.sleep', return_value=None)
    def test_mark_unread(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestInbox.test_mark_unread',
                match_requests_on=['uri', 'method', 'body']):
            self.reddit.inbox.mark_unread(list(self.reddit.inbox.all()))

    def test_messages(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestInbox.test_messages'):
            count = 0
            for item in self.reddit.inbox.messages(limit=64):
                assert isinstance(item, Message)
                count += 1
            assert count == 64

    def test_sent(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestInbox.test_sent'):
            count = 0
            for item in self.reddit.inbox.sent(limit=64):
                assert isinstance(item, Message)
                assert item.author == self.reddit.config.username
                count += 1
            assert count == 64

    def test_submission_replies(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestInbox.test_submission_replies'):
            count = 0
            for item in self.reddit.inbox.submission_replies(limit=64):
                assert isinstance(item, Comment)
                assert item.parent_id.startswith(
                    self.reddit.config.kinds['submission'])
                count += 1
            assert count == 64

    @mock.patch('time.sleep', return_value=None)
    def test_unread(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestInbox.test_unread'):
            count = 0
            for item in self.reddit.inbox.unread(limit=64):
                count += 1
            assert count == 64



"""Provide the RedditorListingMixin class."""
from six.moves.urllib.parse import urljoin  # pylint: disable=import-error

from ..generator import ListingGenerator
from .base import BaseListingMixin
from .gilded import GildedListingMixin


class RedditorListingMixin(BaseListingMixin, GildedListingMixin):
    """Adds additional methods pertaining to Redditor instances."""

    @property
    def comments(self):
        """An attribute representing the comments made by the Redditor."""
        if self.__dict__.get('_comments') is None:
            self._comments = SubListing(self._reddit, self._path, 'comments')
        return self._comments

    @property
    def submissions(self):
        """An attribute representing the submissions made by the Redditor."""
        if self.__dict__.get('_submissions') is None:
            self._submissions = SubListing(self._reddit, self._path,
                                           'submitted')
        return self._submissions

    def downvoted(self, **generator_kwargs):
        """Return a ListingGenerator for items the user has downvoted.

        May raise ``prawcore.Forbidden`` after issuing the request if the user
        is not authorized to access the list. Note that because this function
        returns a ``ListingGenerator`` the exception may not occur until
        sometime after this function has returned.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, urljoin(self._path, 'downvoted'),
                                **generator_kwargs)

    def gildings(self, **generator_kwargs):
        """Return a ListingGenerator for items the user has gilded.

        May raise ``prawcore.Forbidden`` after issuing the request if the user
        is not authorized to access the list. Note that because this function
        returns a ``ListingGenerator`` the exception may not occur until
        sometime after this function has returned.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit,
                                urljoin(self._path, 'gilded/given'),
                                **generator_kwargs)

    def hidden(self, **generator_kwargs):
        """Return a ListingGenerator for items the user has hidden.

        May raise ``prawcore.Forbidden`` after issuing the request if the user
        is not authorized to access the list. Note that because this function
        returns a ``ListingGenerator`` the exception may not occur until
        sometime after this function has returned.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, urljoin(self._path, 'hidden'),
                                **generator_kwargs)

    def saved(self, **generator_kwargs):
        """Return a ListingGenerator for items the user has saved.

        May raise ``prawcore.Forbidden`` after issuing the request if the user
        is not authorized to access the list. Note that because this function
        returns a ``ListingGenerator`` the exception may not occur until
        sometime after this function has returned.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, urljoin(self._path, 'saved'),
                                **generator_kwargs)

    def upvoted(self, **generator_kwargs):
        """Return a ListingGenerator for items the user has upvoted.

        May raise ``prawcore.Forbidden`` after issuing the request if the user
        is not authorized to access the list. Note that because this function
        returns a ``ListingGenerator`` the exception may not occur until
        sometime after this function has returned.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, urljoin(self._path, 'upvoted'),
                                **generator_kwargs)


class SubListing(BaseListingMixin):
    """Helper class for generating SubListing objects."""

    def __init__(self, reddit, base_path, subpath):
        """Initialize a SubListing instance.

        :param reddit: An instance of :class:`.Reddit'.
        :param base_path: The path to the object up to this point.
        :param subpath: The additional path to this sublisting.

        """
        super(SubListing, self).__init__(reddit, None)
        self._listing_use_sort = True
        self._reddit = reddit
        self._path = urljoin(base_path, subpath)
